"""Identification of individual fragments given the predictions generate by the idCNN
"""
import logging
from shutil import copyfile

import numpy as np
from torch import load, nn

from idtrackerai import Fragment, ListOfFragments
from idtrackerai.network import NetworkParams
from idtrackerai.utils import Timer

from .identity_network import get_predictions_identities


def compute_identification_statistics_for_non_accumulated_fragments(
    fragments: list[Fragment],
    all_predictions: np.ndarray,
    all_softmax_probs: np.ndarray,
    number_of_animals: int,
):
    """Given the predictions associated to the images in each (individual)
    fragment in the list fragments if computes the statistics necessary for the
    identification of fragment.

    Parameters
    ----------
    fragments : list
        List of individual fragment objects
    assigner : <GetPrediction object>
        The assigner object has as main attributes the list of predictions
        associated to `images` and the the corresponding softmax vectors
    number_of_animals : int
        number of animals to be tracked
    """
    counter = 0
    for fragment in fragments:
        if not fragment.used_for_training and fragment.is_an_individual:
            next_counter_value = counter + fragment.number_of_images
            predictions = all_predictions[counter:next_counter_value]
            softmax_probs = all_softmax_probs[counter:next_counter_value]
            fragment.compute_identification_statistics(
                predictions, softmax_probs, number_of_animals
            )
            counter = next_counter_value


def check_penultimate_model(
    identification_model: nn.Module, network_params: NetworkParams
):
    """Loads the penultimate accumulation step if the validation accuracy of the last
    step was lower then the penultimate. This discard possible corrupt final accumulation steps
    """
    if not network_params.penultimate_model_path.is_file():
        logging.warning(
            "Penultimate model not found (%s)", network_params.penultimate_model_path
        )
        return

    last_accuracy = load(network_params.model_path).get("val_acc", 0.0)
    penultimate_model: dict = load(network_params.penultimate_model_path)
    penultimate_accuracy = penultimate_model.pop("val_acc", -1.0)
    logging.info(
        f"Last accuracy = {last_accuracy:.2%}, "
        f"Penultimate accuracy = {penultimate_accuracy:.2%}"
    )
    if penultimate_accuracy > last_accuracy:
        logging.info(
            "The last accumulation step had a lower accuracy than the penultimate."
        )
        logging.info(
            "Loading penultimate model, %s", network_params.penultimate_model_path
        )
        identification_model.load_state_dict(penultimate_model, strict=True)

        # set the penultimate as the one model
        network_params.model_path.unlink()
        copyfile(network_params.penultimate_model_path, network_params.model_path)
    else:
        logging.info(
            "The last accumulation step had a higher accuracy than the penultimate."
        )

    network_params.penultimate_model_path.unlink()


def assign_remaining_fragments(
    list_of_fragments: ListOfFragments,
    identification_model: nn.Module,
    network_params: NetworkParams,
    timer: Timer,
):
    """This is the main function of this module: given a list_of_fragments it
    puts in place the routine to identify, if possible, each of the individual
    fragments. The starting point for the identification is given by the
    predictions produced by the ConvNetwork net passed as input. The organisation
    of the images in individual fragments is then used to assign more accurately.

    Parameters
    ----------
    list_of_fragments : <ListOfFragments object>
        collection of the individual fragments and associated methods
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    net : <ConvNetwork object>
        Convolutional neural network object created according to net.params

    See Also
    --------
    ListOfFragments.get_images_from_fragments_to_assign
    assign
    compute_identification_statistics_for_non_accumulated_fragments

    """
    timer.start()

    check_penultimate_model(identification_model, network_params)
    logging.info("Assigning identities to all non-accumulated individual fragments")
    list_of_fragments.reset(roll_back_to="accumulation")
    number_of_unidentified_individual_fragments = (
        list_of_fragments.get_number_of_unidentified_individual_fragments()
    )
    logging.info(
        "Number of unidentified individual fragments: "
        f"{number_of_unidentified_individual_fragments}"
    )
    if not number_of_unidentified_individual_fragments:
        list_of_fragments.compute_P2_vectors()
        return

    images = list_of_fragments.get_images_from_fragments_to_assign()

    predictions, softmax_probs = get_predictions_identities(
        identification_model, images, network_params
    )

    logging.debug(
        f"{len(predictions)} generated predictions between "
        f"identities {set(predictions)}"
    )
    compute_identification_statistics_for_non_accumulated_fragments(
        list_of_fragments.fragments,
        predictions,
        softmax_probs,
        list_of_fragments.n_animals,
    )

    logging.info("Assigning identities")
    list_of_fragments.compute_P2_vectors()
    fragment = list_of_fragments.get_next_fragment_to_identify()
    while fragment:
        fragment.assign_identity(list_of_fragments.n_animals)
        fragment = list_of_fragments.get_next_fragment_to_identify()

    timer.finish()



import logging
import shutil
import sys
from importlib.metadata import version
from importlib.resources import files
from pathlib import Path
from typing import Any

try:
    # PyQt has to be imported before CV2 (importing idtrackerai stuff implies CV2)
    # If not, the QFileDialog.getFileNames() does not load the icons, very weird
    from qtpy.QtWidgets import QApplication
except ImportError:
    logging.error(
        "\n\tRUNNING AN IDTRACKER.AI INSTALLATION WITHOUT ANY QT BINDING.\n\tGUIs are"
        " not available, only tracking directly from the terminal with the `--track`"
        " flag.\n\tRun `pip install pyqt5` or `pip install pyqt6` to build a Qt"
        " binding."
    )

import toml

from idtrackerai import Video
from idtrackerai.utils import CustomError, conf, pprint_dict, wrap_entrypoint

from .arg_parser import parse_args


def load_toml(path: Path, name: str = "") -> dict:
    if not path.is_file():
        raise FileNotFoundError(f"{path} do not exist")
    try:
        toml_dict = {
            key.lower(): value for key, value in toml.load(path.open()).items()
        }

        for key, value in toml_dict.items():
            if value == "":
                toml_dict[key] = None
        if name:
            logging.info(pprint_dict(toml_dict, name), extra={"markup": True})
        return toml_dict
    except Exception as exc:
        raise CustomError(f"Could not read {path}.\n" + str(exc)) from exc


def gather_input_parameters() -> tuple[bool, dict[str, Any]]:
    parameters = {}
    if Path("local_settings.py").is_file():
        logging.warning("Deprecated local_settings format found in ./local_settings.py")

    local_settings_path = Path("local_settings.toml")
    if local_settings_path.is_file():
        parameters = load_toml(local_settings_path, "Local settings")

    terminal_args = parse_args()
    ready_to_track = terminal_args.pop("track")

    if "general_settings" in terminal_args:
        general_settings = load_toml(
            terminal_args.pop("general_settings"), "General settings"
        )
        parameters.update(general_settings)
    else:
        logging.info("No general settings loaded")

    if "session_parameters" in terminal_args:
        session_parameters = load_toml(
            terminal_args.pop("session_parameters"), "Session parameters"
        )
        parameters.update(session_parameters)
    else:
        logging.info("No session parameters loaded")

    if terminal_args:
        logging.info(
            pprint_dict(terminal_args, "Terminal arguments"), extra={"markup": True}
        )
        parameters.update(terminal_args)
    else:
        logging.info("No terminal arguments detected")
    return ready_to_track, parameters


@wrap_entrypoint
def main() -> bool:
    """The command `idtrackerai` runs this function"""
    ready_to_track, user_parameters = gather_input_parameters()

    video = Video()
    non_recognized_params_1 = conf.set_parameters(**user_parameters)
    non_recognized_params_2 = video.set_parameters(**user_parameters)

    non_recognized_params = non_recognized_params_1 & non_recognized_params_2

    if non_recognized_params:
        raise CustomError(f"Not recognized parameters: {non_recognized_params}")

    if not ready_to_track:
        ready_to_track = run_segmentation_GUI(video)
        if not ready_to_track:
            return False

    from .run_idtrackerai import RunIdTrackerAi

    return RunIdTrackerAi(video).track_video()


def run_segmentation_GUI(video: Video | None) -> bool:
    try:
        from idtrackerai_start_app.segmentation_GUI import SegmentationGUI
    except ImportError as exc:
        raise CustomError(
            "\n\tRUNNING AN IDTRACKER.AI INSTALLATION WITHOUT ANY QT BINDING.\n\tGUIs"
            " are not available, only tracking directly from the terminal with the"
            " `--track` flag.\n\tRun `pip install pyqt5` or `pip install pyqt6` to"
            " build a Qt binding."
        ) from exc
    app = QApplication(sys.argv)
    signal = {"run_idtrackerai": False}
    window = SegmentationGUI(video, signal)
    window.show()
    app.exec()
    return signal["run_idtrackerai"] is True


@wrap_entrypoint
def general_test():
    from datetime import datetime

    from .run_idtrackerai import RunIdTrackerAi

    COMPRESSED_VIDEO_PATH = Path(str(files("idtrackerai"))) / "data" / "test_B.avi"

    video_path = Path.cwd() / COMPRESSED_VIDEO_PATH.name
    shutil.copyfile(COMPRESSED_VIDEO_PATH, video_path)

    video = Video()
    video.set_parameters(
        session="test",
        video_paths=video_path,
        tracking_intervals=None,
        intensity_ths=[0, 130],
        area_ths=[150, 60000],
        number_of_animals=8,
        resolution_reduction=1.0,
        check_segmentation=False,
        ROI_list=None,
        track_wo_identities=False,
        use_bkg=False,
        protocol3_action="continue",
    )

    start = datetime.now()
    success = RunIdTrackerAi(video).track_video()
    if success:
        logging.info(
            "[green]Test passed successfully in %s with version %s",
            str(datetime.now() - start).split(".")[0],
            version("idtrackerai"),
            extra={"markup": True},
        )


if __name__ == "__main__":
    main()



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging
from shutil import copyfile

import torch
from torch.backends import cudnn
from torch.nn import CrossEntropyLoss, Module
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import Video
from idtrackerai.network import DEVICE, LearnerClassification, NetworkParams
from idtrackerai.utils import conf

from .accumulation_manager import (
    AccumulationManager,
    get_predictions_of_candidates_fragments,
)
from .identity_dataset import get_training_data_loaders, split_data_train_and_validation
from .identity_network import StopTraining, TrainIdentification


def perform_one_accumulation_step(
    accumulation_manager: AccumulationManager,
    video: Video,
    identification_model: Module,
    network_params: NetworkParams,
):
    logging.info(
        f"[bold]Performing new accumulation, step {accumulation_manager.current_step}",
        extra={"markup": True},
    )

    # Get images for training
    accumulation_manager.get_new_images_and_labels()
    images, labels = accumulation_manager.get_images_and_labels_for_training()
    train_data, val_data = split_data_train_and_validation(
        images, labels, validation_proportion=conf.VALIDATION_PROPORTION
    )
    assert images.shape[0] == labels.shape[0]
    logging.info(
        "Training with %d, validating with %d",
        len(train_data["images"]),
        len(val_data["images"]),
    )
    assert len(val_data["images"]) > 0

    # Set data loaders
    train_loader, val_loader = get_training_data_loaders(
        video.n_animals, train_data, val_data
    )

    criterion = CrossEntropyLoss(weight=torch.tensor(train_data["weights"]))

    logging.info("Sending model and criterion to %s", DEVICE)
    cudnn.benchmark = True  # make it train faster
    identification_model.to(DEVICE)
    criterion.to(DEVICE)

    logging.info(f"Setting {network_params.optimizer} optimizer")
    if network_params.optimizer == "Adam":
        optimizer = torch.optim.Adam(
            identification_model.parameters(), **network_params.optim_args
        )
    elif network_params.optimizer == "SGD":
        optimizer = torch.optim.SGD(
            identification_model.parameters(), **network_params.optim_args
        )
    else:
        raise AttributeError(network_params.optimizer)

    scheduler = MultiStepLR(optimizer, milestones=network_params.schedule, gamma=0.1)

    learner = LearnerClassification(
        identification_model, criterion, optimizer, scheduler
    )

    stop_training = StopTraining(
        network_params.n_classes,
        is_first_accumulation=accumulation_manager.current_step == 0,
    )

    # keep a copy of the penultimate model
    network_params.penultimate_model_path.unlink(missing_ok=True)
    if network_params.model_path.is_file():
        copyfile(network_params.model_path, network_params.penultimate_model_path)

    TrainIdentification(
        learner, train_loader, val_loader, network_params, stop_training
    )

    accumulation_manager.update_fragments_used_for_training()
    accumulation_manager.update_used_images_and_labels()
    accumulation_manager.assign_identities_to_fragments_used_for_training()
    accumulation_manager.update_set_of_individual_fragments_used()

    # compute ratio of accumulated images and stop if it is above random
    accumulation_manager.ratio_accumulated_images = (
        accumulation_manager.list_of_fragments.ratio_of_images_used_for_training
    )

    if (
        accumulation_manager.ratio_accumulated_images
        > conf.THRESHOLD_EARLY_STOP_ACCUMULATION
    ):
        logging.debug("Stopping accumulation by early stopping criteria")
        return

    # Set accumulation parameters for rest of the accumulation
    # take images from global fragments not used in training (in the remainder test global fragments)
    if any(
        not global_fragment.used_for_training
        for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments
    ):
        logging.info(
            "Generating [bold]predictions[/bold] on remaining global fragments",
            extra={"markup": True},
        )
        (
            predictions,
            softmax_probs,
            indices_to_split,
            candidate_individual_fragments_identifiers,
        ) = get_predictions_of_candidates_fragments(
            identification_model,
            video.id_images_file_paths,
            network_params,
            accumulation_manager.list_of_fragments,
        )

        accumulation_manager.split_predictions_after_network_assignment(
            predictions,
            softmax_probs,
            indices_to_split,
            candidate_individual_fragments_identifiers,
        )
        # assign identities to the global fragments based on the predictions
        logging.info(
            "Checking eligibility criteria and generate the "
            "new list of identified global fragments to accumulate"
        )
        accumulation_manager.get_acceptable_global_fragments_for_training(
            candidate_individual_fragments_identifiers, video.accumulation_trial
        )

        accumulation_manager.print_accumulation_variables()

        stats = video.accumulation_statistics

        stats["n_accumulated_global_fragments"].append(
            sum(
                global_fragment.used_for_training
                for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments
            )
        )
        stats["n_non_certain_global_fragments"].append(
            accumulation_manager.number_of_noncertain_global_fragments
        )
        stats["n_randomly_assigned_global_fragments"].append(
            accumulation_manager.number_of_random_assigned_global_fragments
        )
        stats["n_nonconsistent_global_fragments"].append(
            accumulation_manager.number_of_nonconsistent_global_fragments
        )
        stats["n_nonunique_global_fragments"].append(
            accumulation_manager.number_of_nonunique_global_fragments
        )
        stats["n_acceptable_global_fragments"].append(
            sum(
                global_fragment.acceptable_for_training(
                    accumulation_manager.accumulation_strategy
                )
                for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments
            )
        )
        stats["ratio_of_accumulated_images"].append(
            accumulation_manager.ratio_accumulated_images
        )

        accumulation_manager.current_step += 1

    accumulation_manager.ratio_accumulated_images = (
        accumulation_manager.list_of_fragments.ratio_of_images_used_for_training
    )

    while len(video.accumulation_statistics_data) <= video.accumulation_trial:
        video.accumulation_statistics_data.append({})

    video.accumulation_statistics_data[video.accumulation_trial] = (
        video.accumulation_statistics
    )



from qtpy.QtGui import QColor, QColorConstants, QPalette

ColorRole = QPalette.ColorRole
ColorGroup = QPalette.ColorGroup
Disabled = QPalette.ColorGroup.Disabled

Red = QColorConstants.Red
Black = QColorConstants.Black
White = QColorConstants.White
Blue = QColorConstants.Blue

dark = QPalette()
dark.setColor(ColorGroup.Active, ColorRole.WindowText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Active, ColorRole.Button, QColor(0x323232))
dark.setColor(ColorGroup.Active, ColorRole.Light, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Active, ColorRole.Midlight, QColor(0x2A2A2A))
dark.setColor(ColorGroup.Active, ColorRole.Dark, QColor(0x212121))
dark.setColor(ColorGroup.Active, ColorRole.Mid, QColor(0x262626))
dark.setColor(ColorGroup.Active, ColorRole.Text, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Active, ColorRole.BrightText, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Active, ColorRole.ButtonText, QColor(0xAAD4FF))
dark.setColor(ColorGroup.Active, ColorRole.Base, QColor(0x242424))
dark.setColor(ColorGroup.Active, ColorRole.Window, QColor(0x323232))
dark.setColor(ColorGroup.Active, ColorRole.Shadow, QColor(0x191919))
dark.setColor(ColorGroup.Active, ColorRole.Highlight, QColor(0x308CC6))
dark.setColor(ColorGroup.Active, ColorRole.HighlightedText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Active, ColorRole.Link, QColor(0x308CC6))
dark.setColor(ColorGroup.Active, ColorRole.LinkVisited, QColor(0xFF00FF))
dark.setColor(ColorGroup.Active, ColorRole.AlternateBase, QColor(0x2B2B2B))
dark.setColor(ColorGroup.Active, ColorRole.ToolTipBase, QColor(0xFFFFDC))
dark.setColor(ColorGroup.Active, ColorRole.ToolTipText, Black)
dark.setColor(ColorGroup.Active, ColorRole.PlaceholderText, QColor(0xA0A0A0))
dark.setColor(ColorGroup.Active, ColorRole.NoRole, Black)
dark.setColor(ColorGroup.Disabled, ColorRole.WindowText, QColor(0x828282))
dark.setColor(ColorGroup.Disabled, ColorRole.Button, QColor(0x323232))
dark.setColor(ColorGroup.Disabled, ColorRole.Light, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Disabled, ColorRole.Midlight, QColor(0x2A2A2A))
dark.setColor(ColorGroup.Disabled, ColorRole.Dark, QColor(0xBEBEBE))
dark.setColor(ColorGroup.Disabled, ColorRole.Mid, QColor(0x262626))
dark.setColor(ColorGroup.Disabled, ColorRole.Text, QColor(0x828282))
dark.setColor(ColorGroup.Disabled, ColorRole.BrightText, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Disabled, ColorRole.ButtonText, QColor(0x828282))
dark.setColor(ColorGroup.Disabled, ColorRole.Base, QColor(0x323232))
dark.setColor(ColorGroup.Disabled, ColorRole.Window, QColor(0x323232))
dark.setColor(ColorGroup.Disabled, ColorRole.Shadow, QColor(0x252525))
dark.setColor(ColorGroup.Disabled, ColorRole.Highlight, QColor(0x919191))
dark.setColor(ColorGroup.Disabled, ColorRole.HighlightedText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Disabled, ColorRole.Link, QColor(0x308CC6))
dark.setColor(ColorGroup.Disabled, ColorRole.LinkVisited, QColor(0xFF00FF))
dark.setColor(ColorGroup.Disabled, ColorRole.AlternateBase, QColor(0x2B2B2B))
dark.setColor(ColorGroup.Disabled, ColorRole.ToolTipBase, QColor(0xFFFFDC))
dark.setColor(ColorGroup.Disabled, ColorRole.ToolTipText, Black)
dark.setColor(ColorGroup.Disabled, ColorRole.PlaceholderText, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Disabled, ColorRole.NoRole, Black)
dark.setColor(ColorGroup.Inactive, ColorRole.WindowText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.Button, QColor(0x323232))
dark.setColor(ColorGroup.Inactive, ColorRole.Light, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Inactive, ColorRole.Midlight, QColor(0x2A2A2A))
dark.setColor(ColorGroup.Inactive, ColorRole.Dark, QColor(0x212121))
dark.setColor(ColorGroup.Inactive, ColorRole.Mid, QColor(0x262626))
dark.setColor(ColorGroup.Inactive, ColorRole.Text, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.BrightText, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Inactive, ColorRole.ButtonText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.Base, QColor(0x242424))
dark.setColor(ColorGroup.Inactive, ColorRole.Window, QColor(0x323232))
dark.setColor(ColorGroup.Inactive, ColorRole.Shadow, QColor(0x191919))
dark.setColor(ColorGroup.Inactive, ColorRole.Highlight, QColor(0x308CC6))
dark.setColor(ColorGroup.Inactive, ColorRole.HighlightedText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.Link, QColor(0x308CC6))
dark.setColor(ColorGroup.Inactive, ColorRole.LinkVisited, QColor(0xFF00FF))
dark.setColor(ColorGroup.Inactive, ColorRole.AlternateBase, QColor(0x2B2B2B))
dark.setColor(ColorGroup.Inactive, ColorRole.ToolTipBase, QColor(0xFFFFDC))
dark.setColor(ColorGroup.Inactive, ColorRole.ToolTipText, Black)
dark.setColor(ColorGroup.Inactive, ColorRole.PlaceholderText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.NoRole, Black)


light = QPalette()
light.setColor(ColorGroup.Active, ColorRole.WindowText, Black)
light.setColor(ColorGroup.Active, ColorRole.Button, QColor(0xEFEFEF))
light.setColor(ColorGroup.Active, ColorRole.Light, White)
light.setColor(ColorGroup.Active, ColorRole.Midlight, QColor(0xCACACA))
light.setColor(ColorGroup.Active, ColorRole.Dark, QColor(0x9F9F9F))
light.setColor(ColorGroup.Active, ColorRole.Mid, QColor(0xB8B8B8))
light.setColor(ColorGroup.Active, ColorRole.Text, Black)
light.setColor(ColorGroup.Active, ColorRole.BrightText, White)
light.setColor(ColorGroup.Active, ColorRole.ButtonText, Black)
light.setColor(ColorGroup.Active, ColorRole.Base, White)
light.setColor(ColorGroup.Active, ColorRole.Window, QColor(0xEFEFEF))
light.setColor(ColorGroup.Active, ColorRole.Shadow, QColor(0x767676))
light.setColor(ColorGroup.Active, ColorRole.Highlight, QColor(0x308CC6))
light.setColor(ColorGroup.Active, ColorRole.HighlightedText, White)
light.setColor(ColorGroup.Active, ColorRole.Link, Blue)
light.setColor(ColorGroup.Active, ColorRole.LinkVisited, QColor(0xFF00FF))
light.setColor(ColorGroup.Active, ColorRole.AlternateBase, QColor(0xF7F7F7))
light.setColor(ColorGroup.Active, ColorRole.ToolTipBase, QColor(0xFFFFDC))
light.setColor(ColorGroup.Active, ColorRole.ToolTipText, Black)
light.setColor(ColorGroup.Active, ColorRole.PlaceholderText, QColor(0x404040))
light.setColor(ColorGroup.Active, ColorRole.NoRole, Black)
light.setColor(ColorGroup.Disabled, ColorRole.WindowText, QColor(0xBEBEBE))
light.setColor(ColorGroup.Disabled, ColorRole.Button, QColor(0xEFEFEF))
light.setColor(ColorGroup.Disabled, ColorRole.Light, White)
light.setColor(ColorGroup.Disabled, ColorRole.Midlight, QColor(0xCACACA))
light.setColor(ColorGroup.Disabled, ColorRole.Dark, QColor(0xBEBEBE))
light.setColor(ColorGroup.Disabled, ColorRole.Mid, QColor(0xB8B8B8))
light.setColor(ColorGroup.Disabled, ColorRole.Text, QColor(0xBEBEBE))
light.setColor(ColorGroup.Disabled, ColorRole.BrightText, White)
light.setColor(ColorGroup.Disabled, ColorRole.ButtonText, QColor(0xBEBEBE))
light.setColor(ColorGroup.Disabled, ColorRole.Base, QColor(0xEFEFEF))
light.setColor(ColorGroup.Disabled, ColorRole.Window, QColor(0xEFEFEF))
light.setColor(ColorGroup.Disabled, ColorRole.Shadow, QColor(0xB1B1B1))
light.setColor(ColorGroup.Disabled, ColorRole.Highlight, QColor(0x919191))
light.setColor(ColorGroup.Disabled, ColorRole.HighlightedText, White)
light.setColor(ColorGroup.Disabled, ColorRole.Link, Blue)
light.setColor(ColorGroup.Disabled, ColorRole.LinkVisited, QColor(0xFF00FF))
light.setColor(ColorGroup.Disabled, ColorRole.AlternateBase, QColor(0xF7F7F7))
light.setColor(ColorGroup.Disabled, ColorRole.ToolTipBase, QColor(0xFFFFDC))
light.setColor(ColorGroup.Disabled, ColorRole.ToolTipText, Black)
light.setColor(ColorGroup.Disabled, ColorRole.PlaceholderText, QColor(0xC0C0C0))
light.setColor(ColorGroup.Disabled, ColorRole.NoRole, Black)
light.setColor(ColorGroup.Inactive, ColorRole.WindowText, Black)
light.setColor(ColorGroup.Inactive, ColorRole.Button, QColor(0xEFEFEF))
light.setColor(ColorGroup.Inactive, ColorRole.Light, White)
light.setColor(ColorGroup.Inactive, ColorRole.Midlight, QColor(0xCACACA))
light.setColor(ColorGroup.Inactive, ColorRole.Dark, QColor(0x9F9F9F))
light.setColor(ColorGroup.Inactive, ColorRole.Mid, QColor(0xB8B8B8))
light.setColor(ColorGroup.Inactive, ColorRole.Text, Black)
light.setColor(ColorGroup.Inactive, ColorRole.BrightText, White)
light.setColor(ColorGroup.Inactive, ColorRole.ButtonText, Black)
light.setColor(ColorGroup.Inactive, ColorRole.Base, White)
light.setColor(ColorGroup.Inactive, ColorRole.Window, QColor(0xEFEFEF))
light.setColor(ColorGroup.Inactive, ColorRole.Shadow, QColor(0x767676))
light.setColor(ColorGroup.Inactive, ColorRole.Highlight, QColor(0x308CC6))
light.setColor(ColorGroup.Inactive, ColorRole.HighlightedText, White)
light.setColor(ColorGroup.Inactive, ColorRole.Link, Blue)
light.setColor(ColorGroup.Inactive, ColorRole.LinkVisited, QColor(0xFF00FF))
light.setColor(ColorGroup.Inactive, ColorRole.AlternateBase, QColor(0xF7F7F7))
light.setColor(ColorGroup.Inactive, ColorRole.ToolTipBase, QColor(0xFFFFDC))
light.setColor(ColorGroup.Inactive, ColorRole.ToolTipText, Black)
light.setColor(ColorGroup.Inactive, ColorRole.PlaceholderText, Black)
light.setColor(ColorGroup.Inactive, ColorRole.NoRole, Black)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
from typing import Iterable

import cv2
import numpy as np
from scipy.spatial.distance import cdist

from idtrackerai import Blob, ListOfBlobs, ListOfFragments, Video
from idtrackerai.utils import track

from .compute_velocity_model import compute_model_velocity
from .erosion import compute_erosion_disk, get_eroded_blobs


def set_individual_with_identity_0_as_crossings(list_of_fragments: ListOfFragments):
    for fragment in list_of_fragments.individual_fragments:
        if (
            len(fragment.assigned_identities) == 1
            and fragment.assigned_identities[0] == 0
        ):
            fragment.is_an_individual = False
            fragment.forced_crossing = True
            fragment.identity = None
            fragment.identity_corrected_solving_jumps = None


def find_the_gap_interval(
    blobs_in_video: list[list[Blob]],
    possible_identities: set[int],
    gap_start: int,
    list_of_occluded_identities: list[set[int]],
):
    there_are_missing_identities = True
    frame_number = gap_start + 1
    if frame_number < len(blobs_in_video):
        while there_are_missing_identities and 0 < frame_number < len(blobs_in_video):
            blobs_in_frame = blobs_in_video[frame_number]
            occluded_identities_in_frame = list_of_occluded_identities[frame_number]
            missing_identities = get_missing_identities_from_blobs_in_frame(
                possible_identities, blobs_in_frame, occluded_identities_in_frame
            )
            if not missing_identities or frame_number == len(blobs_in_video) - 1:
                there_are_missing_identities = False
            else:
                frame_number += 1
            gap_end = frame_number
    else:
        gap_end = gap_start
    return gap_start, gap_end


def get_blob_by_identity(blobs_in_frame: list[Blob], identity: int) -> Blob | None:
    for blob in blobs_in_frame:
        if identity in blob.final_identities:
            return blob
    return None


def get_candidate_blobs_by_overlapping(
    blob_to_test: Blob, eroded_blobs_in_frame: list[Blob]
) -> list[Blob]:
    overlapping_blobs = [
        blob for blob in eroded_blobs_in_frame if blob_to_test.overlaps_with(blob)
    ]
    return overlapping_blobs if overlapping_blobs else eroded_blobs_in_frame


def get_missing_identities_from_blobs_in_frame(
    possible_identities: set[int],
    blobs_in_frame: list[Blob],
    occluded_identities_in_frame: set[int],
) -> set:
    identities_in_frame = set()
    for blob in blobs_in_frame:
        identities_in_frame.update(blob.final_identities)
    return (possible_identities - identities_in_frame) - occluded_identities_in_frame


def get_candidate_centroid(
    individual_gap_interval: tuple[int, int],
    previous_blob_to_the_gap: Blob,
    next_blob_to_the_gap: Blob,
    identity: int,
    border: str,
) -> tuple[float, float]:
    blobs_for_interpolation = [previous_blob_to_the_gap, next_blob_to_the_gap]
    centroids_to_interpolate = [
        list(blob.final_centroids)[list(blob.final_identities).index(identity)]
        for blob in blobs_for_interpolation
    ]
    centroids_to_interpolate = np.asarray(centroids_to_interpolate).T
    argsort_x = np.argsort(centroids_to_interpolate[0])
    centroids_to_interpolate[0] = centroids_to_interpolate[0][argsort_x]
    centroids_to_interpolate[1] = centroids_to_interpolate[1][argsort_x]
    number_of_points = individual_gap_interval[1] - individual_gap_interval[0] + 1
    x_interp = np.linspace(
        centroids_to_interpolate[0][0],
        centroids_to_interpolate[0][1],
        number_of_points + 1,
    )
    y_interp = np.interp(
        x_interp, centroids_to_interpolate[0], centroids_to_interpolate[1]
    )
    if border == "start" and all(argsort_x == np.asarray([0, 1])):
        return list(zip(x_interp, y_interp))[1]
    if border == "start" and all(argsort_x == np.asarray([1, 0])):
        return list(zip(x_interp, y_interp))[-2]
    if border == "end" and all(argsort_x == np.asarray([0, 1])):
        return list(zip(x_interp, y_interp))[-2]
    if border == "end" and all(argsort_x == np.asarray([1, 0])):
        return list(zip(x_interp, y_interp))[1]
    raise ValueError(border, argsort_x)


def find_the_individual_gap_interval(
    blobs_in_video: list[list[Blob]],
    possible_identities: set[int],
    identity: int,
    a_frame_in_the_gap: int,
    list_of_occluded_identities: list[set[int]],
) -> tuple[int, int]:
    # logging.debug('Finding the individual gap interval')
    # find gap start
    identity_is_missing = True
    gap_start = a_frame_in_the_gap
    frame_number = gap_start

    # logging.debug('To the past while loop')
    while identity_is_missing and 0 < frame_number < len(blobs_in_video):
        blobs_in_frame = blobs_in_video[frame_number]
        occluded_identities_in_frame = list_of_occluded_identities[frame_number]
        missing_identities = get_missing_identities_from_blobs_in_frame(
            possible_identities, blobs_in_frame, occluded_identities_in_frame
        )
        if identity not in missing_identities:
            gap_start = frame_number + 1
            identity_is_missing = False
        else:
            frame_number -= 1

    # find gap end
    identity_is_missing = True
    frame_number = a_frame_in_the_gap
    gap_end = a_frame_in_the_gap

    # logging.debug('To the future while loop')
    while identity_is_missing and 0 < frame_number < len(blobs_in_video):
        blobs_in_frame = blobs_in_video[frame_number]
        occluded_identities_in_frame = list_of_occluded_identities[frame_number]
        missing_identities = get_missing_identities_from_blobs_in_frame(
            possible_identities, blobs_in_frame, occluded_identities_in_frame
        )
        if identity not in missing_identities:
            gap_end = frame_number
            identity_is_missing = False
        else:
            gap_end += 1
            frame_number = gap_end
    # logging.debug('Finished finding the individual gap interval')
    return gap_start, gap_end


def get_previous_and_next_blob_wrt_gap(
    blobs_in_video: list[list[Blob]],
    possible_identities: set[int],
    identity: int,
    frame_number: int,
    list_of_occluded_identities: list[set[int]],
) -> tuple[tuple[int, int], Blob | None, Blob | None]:
    # logging.debug('Finding previons and next blobs to the gap of this identity')
    individual_gap_interval = find_the_individual_gap_interval(
        blobs_in_video,
        possible_identities,
        identity,
        frame_number,
        list_of_occluded_identities,
    )
    if individual_gap_interval[0] != 0:
        previous_blob_to_the_gap = get_blob_by_identity(
            blobs_in_video[individual_gap_interval[0] - 1], identity
        )
    else:
        previous_blob_to_the_gap = None

    if individual_gap_interval[1] != len(blobs_in_video):
        next_blob_to_the_gap = get_blob_by_identity(
            blobs_in_video[individual_gap_interval[1]], identity
        )
    else:
        next_blob_to_the_gap = None

    return individual_gap_interval, previous_blob_to_the_gap, next_blob_to_the_gap


def get_closest_contour_point_to(
    contour: np.ndarray, candidate_centroid: tuple[float, float]
):
    return tuple(contour[np.argmin(cdist([candidate_centroid], contour))])


def get_nearest_eroded_blob_to_candidate_centroid(
    eroded_blobs: list[Blob], candidate_centroid: tuple[float, float]
) -> Blob:
    return min(
        eroded_blobs, key=lambda b: b.distance_from_countour_to(candidate_centroid)
    )


def nearest_candidate_blob_is_near_enough(
    velocity_threshold: float,
    candidate_blob: Blob,
    candidate_centroid: tuple[float, float],
    blob_in_border_frame: Blob,
) -> bool:
    points = [candidate_centroid, blob_in_border_frame.centroid]
    return any(
        candidate_blob.distance_to(point) < velocity_threshold for point in points
    )


def centroid_is_inside_of_any_eroded_blob(
    candidate_eroded_blobs: list[Blob], candidate_centroid: tuple[float, float]
) -> list[Blob]:
    # logging.debug('Checking whether the centroids is inside of a blob')
    candidate_centroid = tuple(map(int, candidate_centroid))
    # logging.debug('Finished whether the centroids is inside of a blob')
    return [
        b
        for b in candidate_eroded_blobs
        if cv2.pointPolygonTest(b.contour, candidate_centroid, False) >= 0
    ]


def evaluate_candidate_blobs_and_centroid(
    velocity_threshold: float,
    candidate_eroded_blobs: list[Blob],
    candidate_centroid: tuple[float, float],
    blob_in_border_frame: Blob,
):
    blob_containing_candidate_centroid = centroid_is_inside_of_any_eroded_blob(
        candidate_eroded_blobs, candidate_centroid
    )
    if blob_containing_candidate_centroid:
        return blob_containing_candidate_centroid[0], candidate_centroid
    if candidate_eroded_blobs:
        nearest_blob = get_nearest_eroded_blob_to_candidate_centroid(
            candidate_eroded_blobs, candidate_centroid
        )
        # TODO why the nearest contour point is the new centroid?
        new_centroid = get_closest_contour_point_to(
            nearest_blob.contour, candidate_centroid
        )
        if nearest_candidate_blob_is_near_enough(
            velocity_threshold, nearest_blob, candidate_centroid, blob_in_border_frame
        ) or nearest_blob.overlaps_with(blob_in_border_frame):
            # the candidate centroid is near to a candidate blob
            return nearest_blob, new_centroid
        # the candidate centroids is far from a candidate blob
        return None, None
    # there where no candidate blobs
    return None, None


def get_candidate_tuples_with_centroids_in_original_blob(
    original_blob: Blob,
    candidate_tuples_to_close_gap: list[tuple[Blob, tuple[float, float], int]],
):
    return [
        candidate_tuple
        for candidate_tuple in candidate_tuples_to_close_gap
        if cv2.pointPolygonTest(
            original_blob.contour, tuple(map(int, candidate_tuple[1])), False
        )
        >= 0
    ]


def assign_identity_to_new_blobs(
    blobs_in_video: list[list[Blob]],
    original_inner_blobs_in_frame: list[Blob],
    candidate_tuples_to_close_gap: list[tuple[Blob, tuple[float, float], int]],
    list_of_occluded_identities: list[set[int]],
):
    new_original_blobs: list[Blob] = []

    for original_blob in original_inner_blobs_in_frame:
        candidate_tuples_with_centroids_in_original_blob = (
            get_candidate_tuples_with_centroids_in_original_blob(
                original_blob, candidate_tuples_to_close_gap
            )
        )
        if len(candidate_tuples_with_centroids_in_original_blob) == 1:
            # the gap is a single individual blob
            identity = candidate_tuples_with_centroids_in_original_blob[0][2]
            centroid = candidate_tuples_with_centroids_in_original_blob[0][1]
            original_blob_final_identities = list(original_blob.final_identities)
            if (
                original_blob.is_an_individual
                and len(original_blob_final_identities) == 1
                and original_blob_final_identities[0] == 0
            ):
                original_blob.identities_corrected_closing_gaps = [identity]
                # TODO loop over the fragment
                for blobs_in_frame in blobs_in_video:
                    for blob in blobs_in_frame:
                        if (
                            blob.fragment_identifier
                            == original_blob.fragment_identifier
                        ):
                            blob.identities_corrected_closing_gaps = [identity]

            elif original_blob.is_an_individual:
                list_of_occluded_identities[original_blob.frame_number].add(identity)
            elif original_blob.is_a_crossing:
                if original_blob.identities_corrected_closing_gaps is not None:
                    identity = original_blob.identities_corrected_closing_gaps + [
                        identity
                    ]
                    assert original_blob.interpolated_centroids is not None
                    centroid = original_blob.interpolated_centroids + [centroid]
                else:
                    identity = [identity]
                    centroid = [centroid]
                frame_number = original_blob.frame_number
                new_blob = candidate_tuples_with_centroids_in_original_blob[0][0]
                new_blob.frame_number = frame_number
                new_blob.identities_corrected_closing_gaps = identity
                new_blob.interpolated_centroids = centroid
                original_blob = new_blob

            new_original_blobs.append(original_blob)
        elif (
            len(candidate_tuples_with_centroids_in_original_blob) > 1
            and original_blob.is_a_crossing
        ):
            # Note that the original blobs that were unidentified (identity 0)
            # are set to zero before starting the main while loop
            # logging.debug('Many candidate tuples for this original blob, '
            #              'and the original blob is a crossing')
            # HERE
            candidate_eroded_blobs: list[Blob] = []
            candidate_eroded_blobs_centroids: list[tuple[float, float]] = []
            candidate_eroded_blobs_identities: list[int] = []
            for blob, centroid, id in candidate_tuples_with_centroids_in_original_blob:
                candidate_eroded_blobs.append(blob)
                candidate_eroded_blobs_centroids.append(centroid)
                candidate_eroded_blobs_identities.append(id)

            if len(set(candidate_eroded_blobs)) == 1:  # crossing not split
                original_blob.interpolated_centroids = candidate_eroded_blobs_centroids
                original_blob.identities_corrected_closing_gaps = (
                    candidate_eroded_blobs_identities
                )

                original_blob.contour = candidate_eroded_blobs[0].contour
                new_original_blobs.append(original_blob)

            elif len(set(candidate_eroded_blobs)) > 1:  # crossing split
                count_eroded_blobs = {
                    eroded_blob: candidate_eroded_blobs.count(eroded_blob)
                    for eroded_blob in candidate_eroded_blobs
                }
                for (
                    eroded_blob,
                    centroid,
                    identity,
                ) in candidate_tuples_with_centroids_in_original_blob:
                    if count_eroded_blobs[eroded_blob] == 1:
                        # split blob, single individual
                        eroded_blob.frame_number = original_blob.frame_number
                        eroded_blob.centroid = centroid
                        eroded_blob.identities_corrected_closing_gaps = [identity]
                        eroded_blob.is_an_individual = True
                        eroded_blob.forced_crossing = True
                        eroded_blob.was_a_crossing = True
                        new_original_blobs.append(eroded_blob)
                    elif count_eroded_blobs[eroded_blob] > 1:
                        if eroded_blob.interpolated_centroids is None:
                            eroded_blob.interpolated_centroids = []
                        if eroded_blob.identities_corrected_closing_gaps is None:
                            eroded_blob.identities_corrected_closing_gaps = []
                        eroded_blob.frame_number = original_blob.frame_number
                        eroded_blob.interpolated_centroids.append(centroid)
                        eroded_blob.identities_corrected_closing_gaps.append(identity)
                        eroded_blob.is_an_individual = False
                        eroded_blob.forced_crossing = True
                        new_original_blobs.append(eroded_blob)

        new_original_blobs.append(original_blob)

    new_original_blobs = list(set(new_original_blobs))
    blobs_in_video[original_blob.frame_number] = new_original_blobs
    return blobs_in_video, list_of_occluded_identities


def get_forward_backward_list_of_frames(gap_interval: tuple[int, int]):
    """input:
    gap_interval: array of tuple [start_frame_number, end_frame_number]
    output:
    [f1, fn, f2, fn-1, ...] for f1 = start_frame_number and
                                fn = end_frame_number"""
    # logging.debug('Got forward-backward list of frames')
    gap_range = range(*gap_interval)
    gap_length = len(gap_range)
    return np.insert(gap_range[::-1], np.arange(gap_length), gap_range)[:gap_length]


def interpolate_trajectories_during_gaps(
    video: Video,
    blobs_in_video: list[list[Blob]],
    list_of_occluded_identities: list[set[int]],
    possible_identities: set[int],
    erosion_counter: int,
):
    for frame_number in track(
        range(1, video.number_of_frames), f"Closing gaps, iteration {erosion_counter}"
    ):
        blobs_in_frame = blobs_in_video[frame_number]
        occluded_identities_in_frame = list_of_occluded_identities[frame_number]

        missing_identities = get_missing_identities_from_blobs_in_frame(
            possible_identities, blobs_in_frame, occluded_identities_in_frame
        )
        if not missing_identities or not blobs_in_frame:
            continue

        gap_interval = find_the_gap_interval(
            blobs_in_video,
            possible_identities,
            frame_number,
            list_of_occluded_identities,
        )
        forward_backward_list_of_frames = get_forward_backward_list_of_frames(
            gap_interval
        )
        for index, inner_frame_number in enumerate(forward_backward_list_of_frames):
            inner_occluded_identities_in_frame = list_of_occluded_identities[
                inner_frame_number
            ]
            inner_blobs_in_frame = blobs_in_video[inner_frame_number]
            if not inner_blobs_in_frame:
                continue

            if erosion_counter == 0:
                eroded_blobs_in_frame = inner_blobs_in_frame
            else:
                eroded_blobs_in_frame = get_eroded_blobs(
                    video, inner_blobs_in_frame, inner_frame_number
                )
                if not eroded_blobs_in_frame:
                    eroded_blobs_in_frame = inner_blobs_in_frame

            inner_missing_identities = get_missing_identities_from_blobs_in_frame(
                possible_identities,
                inner_blobs_in_frame,
                inner_occluded_identities_in_frame,
            )
            candidate_tuples_to_close_gap: list[
                tuple[Blob, tuple[float, float], int]
            ] = []
            for identity in inner_missing_identities:
                (
                    individual_gap_interval,
                    previous_blob_to_the_gap,
                    next_blob_to_the_gap,
                ) = get_previous_and_next_blob_wrt_gap(
                    blobs_in_video,
                    possible_identities,
                    identity,
                    inner_frame_number,
                    list_of_occluded_identities,
                )
                if (
                    previous_blob_to_the_gap is not None
                    and next_blob_to_the_gap is not None
                ):
                    border = "start" if index % 2 == 0 else "end"
                    candidate_centroid = get_candidate_centroid(
                        individual_gap_interval,
                        previous_blob_to_the_gap,
                        next_blob_to_the_gap,
                        identity,
                        border=border,
                    )

                    blob_in_border_frame = (
                        previous_blob_to_the_gap
                        if border == "start"
                        else next_blob_to_the_gap
                    )

                    candidate_eroded_blobs = get_candidate_blobs_by_overlapping(
                        blob_in_border_frame, eroded_blobs_in_frame
                    ) + centroid_is_inside_of_any_eroded_blob(
                        eroded_blobs_in_frame, candidate_centroid
                    )

                    candidate_blob_to_close_gap, centroid = (
                        evaluate_candidate_blobs_and_centroid(
                            video.velocity_threshold,
                            candidate_eroded_blobs,
                            candidate_centroid,
                            blob_in_border_frame,
                        )
                    )
                    if candidate_blob_to_close_gap is not None:
                        assert centroid is not None
                        candidate_tuples_to_close_gap.append(
                            (candidate_blob_to_close_gap, centroid, identity)
                        )
                    else:
                        list_of_occluded_identities[inner_frame_number].add(identity)
                # this manages the case in which identities are
                # missing in the first frame or disappear
                # without appearing anymore,
                else:
                    # and eventual occlusions (an identified blob
                    # does not appear in the previous and/or the
                    # next frame)
                    # logging.debug('------There is not next or not '
                    #              'previous blob to this inner gap:'
                    #              ' it must be occluded')
                    # if previous_blob_to_the_gap is None:
                    #     logging.debug('previous_blob_to_the_gap '
                    #                  'is None')
                    # else:
                    #     logging.debug('previous_blob_to_the_gap '
                    #                  'exists')
                    # if next_blob_to_the_gap:
                    #     logging.debug('next_blob_to_the_gap is None')
                    # else:
                    #     logging.debug('next_blob_to_the_gap exists')
                    for i in range(
                        individual_gap_interval[0], individual_gap_interval[1]
                    ):
                        list_of_occluded_identities[i].add(identity)

            blobs_in_video, list_of_occluded_identities = assign_identity_to_new_blobs(
                blobs_in_video,
                inner_blobs_in_frame,
                candidate_tuples_to_close_gap,
                list_of_occluded_identities,
            )
    return blobs_in_video, list_of_occluded_identities


def reset_blobs_in_video_before_erosion_iteration(all_blobs: Iterable[Blob]):
    """Resets the identity of crossings and individual with multiple identities
    before starting a loop of interpolation

    Parameters
    ----------
    blobs_in_video : list of lists of `Blob` objects
    """
    for blob in all_blobs:
        if blob.is_a_crossing:
            blob.identity = None
        elif blob.is_an_individual and len(list(blob.final_identities)) > 1:
            blob.identities_corrected_closing_gaps = None


def close_trajectories_gaps(
    video: Video, list_of_blobs: ListOfBlobs, list_of_fragments: ListOfFragments
):
    """This is the main function to close the gaps where animals have not been
    identified (labelled with identity 0), are crossing with another animals or
    are occluded or not segmented.

    Parameters
    ----------
    video : <Video object>
        Object containing all the parameters of the video.
    list_of_blobs : <ListOfBlobs object>
        Object with the collection of blobs found during segmentation with associated
        methods. See :class:`list_of_blobs.ListOfBlobs`
    list_of_fragments : <ListOfFragments object>
        Collection of individual and crossing fragments with associated methods.
        See :class:`list_of_fragments.ListOfFragments`

    Returns
    -------
    list_of_blobs : <ListOfBlobs object>
        ListOfBlobs object with the updated blobs and identities that close gaps

    See Also
    --------
    :func:`set_individual_with_identity_0_as_crossings`
    :func:`compute_erosion_disk`
    :func:`compute_model_velocity`
    :func:`reset_blobs_in_video_before_erosion_iteration`
    :func:`interpolate_trajectories_during_gaps`
    :func:`closing_gap_stopping_criteria`
    :func:`clean_individual_blob_before_saving`

    """
    set_individual_with_identity_0_as_crossings(list_of_fragments)
    list_of_fragments.update_blobs(list_of_blobs.all_blobs)

    if not hasattr(video, "erosion_kernel_size"):
        video.erosion_kernel_size = compute_erosion_disk(list_of_blobs.blobs_in_video)
    if not hasattr(video, "velocity_threshold"):
        video.velocity_threshold = compute_model_velocity(list_of_fragments)
    possible_identities = set(range(1, video.n_animals + 1))
    list_of_occluded_identities: list[set[int]] = [
        set() for _ in range(video.number_of_frames)
    ]

    previous_number_of_non_split_crossings = list_of_blobs.number_of_crossing_blobs
    erosion_counter = 0
    continue_erosion_protocol = True
    # TODO why erosion_counter==1?
    while continue_erosion_protocol or erosion_counter == 1:
        reset_blobs_in_video_before_erosion_iteration(list_of_blobs.all_blobs)
        list_of_blobs.blobs_in_video, list_of_occluded_identities = (
            interpolate_trajectories_during_gaps(
                video,
                list_of_blobs.blobs_in_video,
                list_of_occluded_identities,
                possible_identities,
                erosion_counter,
            )
        )

        current_number_of_non_split_crossings = list_of_blobs.number_of_crossing_blobs

        continue_erosion_protocol = (
            previous_number_of_non_split_crossings
            > current_number_of_non_split_crossings
        )

        previous_number_of_non_split_crossings = current_number_of_non_split_crossings
        erosion_counter += 1

    for blob in list_of_blobs.all_blobs:
        if blob.is_an_individual and len(list(blob.final_identities)) > 1:
            blob.identities_corrected_closing_gaps = None
    return list_of_blobs



import json
import logging
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Literal, Optional

from idtrackerai.utils import create_dir, json_default


@dataclass(slots=True)
class NetworkParams:
    n_classes: int
    schedule: list[int]
    architecture: str
    model_name: str
    image_size: list[int]
    optim_args: dict = field(default_factory=dict)
    epochs: int = 0
    optimizer: Literal["Adam", "SGD"] = "SGD"
    loss: str = "CE"
    save_folder: Path = Path("")
    scopes_layers_to_optimize: Optional[list[str]] = field(default_factory=list)
    knowledge_transfer_folder: Path | None = None
    use_adam_optimiser: bool = False
    restore_folder: Path = Path()

    @property
    def load_model_path(self) -> Path:
        # v5.0.0 compatibility
        for deprecated_name in (
            self.model_name + "_.model.pth",
            "supervised_" + self.model_name + ".model.pth",
            "supervised_" + self.model_name + "_.model.pth",
        ):
            path = self.restore_folder / deprecated_name
            if path.is_file():
                return path

        return self.restore_folder / (self.model_name + ".model.pth")

    @property
    def model_path(self) -> Path:
        return (self.save_folder / self.model_name).with_suffix(".model.pth")

    @property
    def penultimate_model_path(self) -> Path:
        return (self.save_folder / (self.model_name + "_penultimate")).with_suffix(
            ".model.pth"
        )

    @property
    def knowledge_transfer_model_file(self) -> Path | None:
        if self.knowledge_transfer_folder is None:
            return None
        return self.knowledge_transfer_folder / "identification_network.model.pth"

    def save(self) -> None:
        path = self.save_folder / "model_params.json"
        logging.info(f"Saving NetworkParams at {path}")
        create_dir(self.save_folder)
        json.dump(asdict(self), path.open("w"), indent=4, default=json_default)



import json
from contextlib import suppress
from pathlib import Path
from time import perf_counter

import numpy as np
import toml
from qtpy.QtCore import QEvent, QRectF, QSize, Qt, QTimer, Signal  # type: ignore
from qtpy.QtGui import (
    QAction,
    QCloseEvent,
    QColor,
    QColorConstants,
    QIcon,
    QImage,
    QKeyEvent,
    QPainter,
    QPixmap,
    QPolygon,
)
from qtpy.QtWidgets import (
    QDialog,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QSlider,
    QSpinBox,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from .canvas import Canvas
from .video_paths_holder import VideoPathHolder


def play_pixmap(size: int):
    canvas = QPixmap(size, size)
    canvas.fill(Qt.GlobalColor.transparent)
    painter = QPainter(canvas)
    pen = painter.pen()
    pen.setColor(QColor(0x306F00))
    pen.setWidth(2)
    painter.setBrush(QColor(0xC0DF50))
    painter.setPen(pen)
    poly = QPolygon()
    poly.setPoints(0, 0, 0, size, size, size // 2)
    painter.drawPolygon(poly)
    return canvas


def pause_pixmap(size: int):
    canvas = QPixmap(size, size)
    canvas.fill(Qt.GlobalColor.transparent)
    painter = QPainter(canvas)
    a = size // 3
    poly = QPolygon()
    pen = painter.pen()
    pen.setColor(QColor(0x404F40))
    pen.setWidth(2)
    painter.setBrush(QColor(0x809F70))
    painter.setPen(pen)
    poly.setPoints(0, 0, a, 0, a, size, 0, size)
    painter.drawPolygon(poly)
    poly.setPoints(size - a, 0, size, 0, size, size, size - a, size)
    painter.drawPolygon(poly)
    return canvas


class VideoPlayer(QWidget):
    painting_time = Signal(QPainter, int, np.ndarray)
    control_bar_h = 30

    def __init__(self, parent: QMainWindow):
        super().__init__(parent)
        self.canvas = Canvas(self)
        self.video_path_holder = VideoPathHolder()

        self.frame_slider = QSlider(Qt.Orientation.Horizontal)
        self.frame_indicator = QSpinBox()

        self.frame_slider.valueChanged.connect(self.frame_indicator.setValue)
        self.frame_slider.sliderPressed.connect(self.stop_all)

        self.frame_indicator.valueChanged.connect(self.frame_indicator_changed)
        self.frame_indicator.setKeyboardTracking(False)
        self.frame_indicator.editingFinished.connect(self.frame_indicator.clearFocus)

        self.time_indicator_widget = QLabel()
        self.play_pause_button = QToolButton()
        self.play_pause_button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.play_pause_button.setShortcut(Qt.Key.Key_Space)
        self.play_pause_button.setCheckable(True)

        icon = QIcon()
        icon.addPixmap(play_pixmap(60), QIcon.Mode.Normal, QIcon.State.Off)
        icon.addPixmap(pause_pixmap(60), QIcon.Mode.Normal, QIcon.State.On)

        self.play_pause_button.setIcon(icon)
        self.play_pause_button.toggled.connect(self.play_pause_clicked)
        self.time_indicator_widget.setFixedHeight(self.control_bar_h)
        self.frame_slider.setFixedHeight(self.control_bar_h)

        self.control_bar = QHBoxLayout()
        self.control_bar.addWidget(self.play_pause_button)
        self.control_bar.addWidget(self.frame_indicator)
        self.control_bar.addWidget(self.frame_slider)
        self.control_bar.addWidget(self.time_indicator_widget)

        layout = QVBoxLayout(self)
        layout.addWidget(self.canvas)
        layout.addLayout(self.control_bar)
        self.time = 0
        self.play_loop = QTimer()
        self.forward_timer = QTimer()
        self.backward_timer = QTimer()
        self.play_loop.timeout.connect(self.forward_loop)
        self.forward_timer.timeout.connect(self.forward_loop)
        self.backward_timer.timeout.connect(self.backward_loop)
        self.min_time_between_frames = 1
        self.fps = 1
        self.drawn_frame = -1
        self.speed: int = 1
        self.speed_label: str = ""
        self.speed_label_timer = QTimer()
        self.speed_label_timer.setSingleShot(True)
        self.speed_label_timer.setInterval(3000)
        self.speed_label_timer.timeout.connect(self.removeSpeedlabel)

        self.freeze = False
        self.canvas.painting_time.connect(self.paint_video)

        menu = parent.menuBar().addMenu("Video player")

        self.draw_in_color = QAction("Enable color", self)
        self.draw_in_color.setCheckable(True)
        menu.addAction(self.draw_in_color)
        self.draw_in_color.toggled.connect(self.update)

        self.limit_framerate = QAction("Limit framerate", self)
        self.limit_framerate.setShortcut("Ctrl+L")
        self.limit_framerate.setCheckable(True)
        menu.addAction(self.limit_framerate)

        self.reduce_cache = QAction("Reduce memory usage", self)
        self.reduce_cache.setCheckable(True)
        menu.addAction(self.reduce_cache)
        self.VideoPlayer_param_path = Path(__file__).parent / "video_player.json"
        self.reduce_cache.toggled.connect(self.video_path_holder.set_cache_mode)
        self.reduce_cache.setChecked(
            json.load(self.VideoPlayer_param_path.open())["reduce_cache"]
            if self.VideoPlayer_param_path.is_file()
            else False
        )

        playback_speed_action = QAction("Change playback speed", self)
        playback_speed_action.triggered.connect(lambda: ChangePlaybackSpeed(self, self.speed))  # type: ignore
        menu.addAction(playback_speed_action)

        def limit_framerate_toggled(state: bool):
            self.min_time_between_frames = 1 / self.fps if state else 0

        self.limit_framerate.toggled.connect(limit_framerate_toggled)

        tooltips = toml.load((Path(__file__).parent.parent / "tooltips.toml"))
        self.draw_in_color.setToolTip(tooltips["color_action"])
        self.limit_framerate.setToolTip(tooltips["framerate_action"])
        self.reduce_cache.setToolTip(tooltips["reducecache_action"])
        menu.setToolTipsVisible(True)
        parent.installEventFilter(self)

    def preload_frames(self, start: int, end: int):
        """Preloads the frames in the video_path_holder cache"""
        color = self.draw_in_color.isChecked()
        for frame in range(start, end):
            self.video_path_holder.frame(frame, color)

    def resizeEvent(self, a0):
        super().resizeEvent(a0)
        play_btn_size = self.frame_indicator.height()
        play_icon_size = int(play_btn_size * 0.6 + 0.5)
        self.play_pause_button.setFixedSize(play_btn_size, play_btn_size)
        self.play_pause_button.setIconSize(QSize(play_icon_size, play_icon_size))

    def closeEvent(self, event: QCloseEvent):
        json.dump(
            {"reduce_cache": self.reduce_cache.isChecked()},
            self.VideoPlayer_param_path.open("w"),
        )
        super().closeEvent(event)

    def event(self, event: QEvent) -> bool:
        if event.type() == QEvent.Type.WindowBlocked:
            self.stop_all()
        return super().event(event)

    def stop_all(self):
        self.play_pause_button.setChecked(False)
        self.forward_timer.stop()
        self.backward_timer.stop()

    def play_pause_clicked(self, play: bool):
        self.forward_timer.stop()
        self.backward_timer.stop()
        if play:
            self.play_loop.start()
        else:
            self.play_loop.stop()

    def frame_indicator_changed(self, frame_indicator_value):
        self.frame_slider.setValue(frame_indicator_value)
        self.update()

    def setCurrentFrame(self, frame, force_update=False):
        self.stop_all()
        if force_update and self.frame_indicator.value() == frame:
            self.update()
        self.frame_indicator.setValue(frame)

    @property
    def current_frame(self) -> int:
        return self.frame_indicator.value()

    @property
    def current_time(self) -> str:
        seconds = int(self.current_frame / self.fps)
        minutes = (seconds // 60) % 60
        hours = (seconds // 3600) % 60
        return f"{hours:02d}:{minutes:02d}:{seconds%60:02d}"

    def paint_video(self, painter: QPainter):
        if not self.isEnabled():
            return

        current_frame = self.current_frame
        self.time_indicator_widget.setText(self.current_time)
        color = self.draw_in_color.isChecked()
        frame = self.video_path_holder.frame(current_frame, color)

        painter.drawImage(
            self.rect_to_draw_image,
            QImage(
                frame.data,
                frame.shape[1],
                frame.shape[0],
                (frame.shape[1] * 3 if color else frame.shape[1]),
                (
                    QImage.Format.Format_BGR888
                    if color
                    else QImage.Format.Format_Grayscale8
                ),
            ),
        )
        self.painting_time.emit(painter, current_frame, frame)

        painter.resetTransform()
        painter.setFont(self.font())
        painter.setPen(QColorConstants.White)
        if self.speed_label:
            painter.drawText(
                self.canvas.rect(), Qt.AlignmentFlag.AlignBottom, self.speed_label
            )

        self.drawn_frame = current_frame

    def pass_frame(self):
        if not self.isEnabled():
            return True
        if self.freeze:
            self.time = perf_counter() + 0.2
            self.freeze = False
            return False
        elapsed_time = perf_counter() - self.time
        if elapsed_time < self.min_time_between_frames:
            return True

        # print(f"  {1/elapsed_time:.4f} fps", end="\r")
        self.time = perf_counter()
        return False

    def backward_loop(self):
        if self.pass_frame():
            return
        new_frame = self.current_frame - self.speed
        if new_frame < 0:
            new_frame = self.n_frames
        self.frame_indicator.setValue(new_frame)

    def forward_loop(self):
        if self.pass_frame():
            return
        new_frame = self.current_frame + self.speed
        if new_frame >= self.n_frames:
            new_frame = 0
        self.frame_indicator.setValue(new_frame)

    def eventFilter(self, object, event: QEvent) -> bool:
        """Catch key events even when VideoPlayer is not in focus."""
        if event.type() == QEvent.Type.KeyPress:
            self.keyPressEvent(event)  # type: ignore
        if event.type() == QEvent.Type.KeyRelease:
            self.keyReleaseEvent(event)  # type: ignore
        return False  # keep processing the event

    def keyPressEvent(self, event: QKeyEvent):
        if event.isAutoRepeat():
            return
        key = event.key()
        if key in (Qt.Key.Key_D, Qt.Key.Key_Right):
            self.freeze = True
            self.forward_timer.start()
            self.play_pause_button.setChecked(False)
        elif key in (Qt.Key.Key_A, Qt.Key.Key_Left):
            self.freeze = True
            self.backward_timer.start()
            self.play_pause_button.setChecked(False)
        with suppress(ValueError):
            self.setSpeed(int(event.text()))
        event.ignore()

    def keyReleaseEvent(self, event: QKeyEvent):
        if event.isAutoRepeat():
            return
        key = event.key()
        if key in (Qt.Key.Key_D, Qt.Key.Key_Right):
            self.forward_timer.stop()
        elif key in (Qt.Key.Key_A, Qt.Key.Key_Left):
            self.backward_timer.stop()

    def setSpeed(self, value: int):
        if value == 0:
            return
        self.speed = 2 ** (value - 1)
        self.speed_label = f"Speed x{self.speed}"
        self.speed_label_timer.start()
        self.update()

    def removeSpeedlabel(self):
        self.speed_label = ""
        self.update()

    def update_video_paths(
        self, video_paths, n_frames, video_size, fps, res_reduct=1.0
    ):
        self.fps = fps
        self.min_time_between_frames = (
            1 / fps if self.limit_framerate.isChecked() else 0
        )
        self.n_frames = n_frames
        self.video_width, self.video_height = video_size
        self.video_path_holder.load_paths(video_paths)
        self.frame_slider.setMaximum(n_frames - 1)
        self.frame_indicator.setMaximum(n_frames - 1)
        self.frame_indicator.setValue(0)
        self.canvas.adjust_zoom_to(
            res_reduct * video_size[0], res_reduct * video_size[1]
        )
        self.set_resolution_reduction(res_reduct)
        self.update()

    def set_resolution_reduction(self, value: float):
        if not hasattr(self, "video_height"):
            return
        self.rect_to_draw_image = QRectF(
            -0.5, -0.5, value * self.video_width, value * self.video_height
        )

    def reorder_video_paths(self, video_paths):
        self.video_path_holder.load_paths(video_paths)
        self.update()

    def center_canvas_at(self, x: float, y: float, zoom_scale: float):
        self.canvas.zoom = zoom_scale / min(self.width(), self.height())
        self.canvas.centerX = x
        self.canvas.centerY = y


class ChangePlaybackSpeed(QDialog):
    def __init__(self, parent: VideoPlayer, current: int):
        super().__init__(parent)
        self.setWindowFlags(Qt.WindowType.Popup)
        self.setLayout(QVBoxLayout())
        self.slider = QSlider(Qt.Orientation.Horizontal)
        self.layout().addWidget(self.slider)
        self.layout().addWidget(QLabel("Change the speed by pressing a numeric key"))
        self.slider.setMinimum(1)
        self.slider.setMaximum(9)
        self.slider.setValue(int(np.log2(current)) + 1)
        self.slider.valueChanged.connect(parent.setSpeed)
        self.exec()

    def keyPressEvent(self, event: QKeyEvent):
        with suppress(ValueError):
            self.slider.setValue(int(event.text()))



"""Given two list of blobs_in_video, one deduced from human ground_truth and the
other generated by the tracking algorithm, compares them and gives back some
statistics

Crossing: crossings are a special case. We ...
"""
import logging
import os
from pprint import pprint

import numpy as np

from idtrackerai import ListOfBlobs, Video


def get_corresponding_gt_blob(blob, gt_blobs_in_frame):
    """Returs the blobs in the `gt_blobs_in_frame` that overlap (in pixels)
    with a given `blob` of the tracked trajectories.

    :param blob: <Blob object>
    :param gt_blobs_in_frame: list of <GroundTruthBlob objects> corresponding
    to the same frame
    :return: list of <GroundTruthBlob objects> that overlap in pixels with
    `blob`
    """
    blobs_of_same_frame = [
        blob.frame_number == gt_blob.frame_number for gt_blob in gt_blobs_in_frame
    ]
    assert all(blobs_of_same_frame)
    corresponding_gt_blobs = []
    for gt_blob in gt_blobs_in_frame:
        if set(gt_blob.pixels).intersection(set(blob.pixels)):
            corresponding_gt_blobs.append(gt_blob)
    return corresponding_gt_blobs


def update_sum_indiv_P2(gt_id, blob, results):
    identified = blob.assigned_identities[0] != 0
    # we only consider P2 for non interpolated blobs
    id_before_interp = blob.identities_corrected_closing_gaps is None
    if identified and id_before_interp:
        results["sum_indiv_P2"][gt_id] += blob._P2_vector[gt_id - 1]


def update_results_with_id_error(results, blob, gt_id):
    results["errors_blobs"][gt_id] += 1
    results["frames_w_id_errors"] = results["frames_w_id_errors"].union(
        {blob.frame_number}
    )
    if blob.fragment_identifier:
        results["frag_w_id_errors"] = results["frag_w_id_errors"].union(
            {blob.fragment_identifier}
        )

    if not blob.used_for_training:
        results["errors_blobs_after_accum"][gt_id] += 1

    # Errors of identified blobs
    if blob.assigned_identities[0] != 0:
        results["errors_id_blobs"][gt_id] += 1

        if blob.used_for_training:
            results["errors_id_blobs_accum"][gt_id] += 1
        else:
            results["errors_id_blobs_after_accum"][gt_id] += 1


def update_results_for_identified_gt_blob(results, blob, gt_id):
    update_sum_indiv_P2(gt_id, blob, results)

    # Count blobs
    results["num_blobs"][gt_id] += 1
    if blob.used_for_training:
        results["num_id_blobs_accum"][gt_id] += 1
    else:
        results["num_blobs_after_accum"][gt_id] += 1

    if blob.assigned_identities[0] != 0:
        results["num_id_blobs"][gt_id] += 1

    if gt_id != blob.assigned_identities[0]:
        update_results_with_id_error(results, blob, gt_id)


def compare_blob_with_gt_blob(results, blob, gt_blob, ids_perm_dict):
    if ids_perm_dict is not None:
        gt_id = ids_perm_dict[gt_blob.gt_identity]
    else:
        gt_id = gt_blob.gt_identity

    if gt_id == 0:
        # This is here to raise and error at the end of the computations
        # A ground truth individual blob cannot have identity 0.
        results["frames_w_0_id_in_gt"] = results["frames_w_0_id_in_gt"].union(
            {gt_blob.frame_number}
        )
    else:
        update_results_for_identified_gt_blob(results, blob, gt_id)


def compare_frame(results, blobs_in_frame, gt_blobs_in_frame, ids_perm_dict):
    for blob in blobs_in_frame:
        corresponding_gt_blobs = get_corresponding_gt_blob(blob, gt_blobs_in_frame)
        if len(corresponding_gt_blobs) == 1:
            gt_blob = corresponding_gt_blobs[0]
            cond1 = gt_blob.is_an_individual
            cond2 = gt_blob.gt_identity != -1
            cond3 = not gt_blob.was_a_crossing
            gt_blob_is_individual = cond1 and cond2 and cond3
            if blob.is_an_individual and gt_blob_is_individual:
                results["num_indiv_gt_blobs"] += 1
                results["num_indiv_blobs"] += 1
                results["crossing_detector_tn"] += 1
                compare_blob_with_gt_blob(results, blob, gt_blob, ids_perm_dict)
            elif blob.is_an_individual and not gt_blob_is_individual:
                # ground truth crossing blob. This could mean that the new
                # video has a better segmentation and it would not be a
                # crossing detection error.
                results["num_crossing_gt_blobs"] += 1
                results["num_indiv_blobs"] += 1
                results["crossing_detector_fn"] += 1
                results["frame_with_crossing_detection_error"] = results[
                    "frame_with_crossing_detection_error"
                ].union({blob.frame_number})
            elif blob.is_a_crossing and gt_blob_is_individual:
                results["num_indiv_gt_blobs"] += 1
                results["num_crossing_blobs"] += 1
                results["crossing_detector_fp"] += 1
                results["frame_with_crossing_detection_error"] = results[
                    "frame_with_crossing_detection_error"
                ].union({blob.frame_number})
            elif blob.is_a_crossing and not gt_blob_is_individual:
                results["num_crossing_gt_blobs"] += 1
                results["num_crossing_blobs"] += 1
                results["crossing_detector_tp"] += 1
        else:
            if blob.is_an_individual:
                results["num_indiv_blobs"] += 1
            else:
                results["num_crossing_blobs"] += 1

    return results


def per_id_counter_dict(num_animals):
    return {i: 0 for i in range(1, num_animals + 1)}


def init_results_dict(num_animals):
    return {
        "num_blobs": per_id_counter_dict(num_animals),
        "sum_indiv_P2": per_id_counter_dict(num_animals),
        "num_id_blobs": per_id_counter_dict(num_animals),
        "num_id_blobs_accum": per_id_counter_dict(num_animals),
        "num_blobs_after_accum": per_id_counter_dict(num_animals),
        "errors_blobs": per_id_counter_dict(num_animals),
        "errors_id_blobs": per_id_counter_dict(num_animals),
        "errors_blobs_after_accum": per_id_counter_dict(num_animals),
        "errors_id_blobs_accum": per_id_counter_dict(num_animals),
        "errors_id_blobs_after_accum": per_id_counter_dict(num_animals),
        "frames_w_id_errors": set(),
        "frag_w_id_errors": set(),
        "frames_w_crossing_errors": set(),
        "frag_w_crossing_errors": set(),
        "frames_w_0_id_in_gt": set(),
        "num_crossing_frags": 0,
        "frag_crossings": set(),
        "crossing_detector_tn": 0,
        "crossing_detector_fn": 0,
        "crossing_detector_tp": 0,
        "crossing_detector_fp": 0,
        "num_indiv_gt_blobs": 0,
        "num_crossing_gt_blobs": 0,
        "num_indiv_blobs": 0,
        "num_crossing_blobs": 0,
        "frame_with_crossing_detection_error": set(),
    }


def aggregate_counters(results):
    results["total_sum_P2"] = np.sum(list(results["sum_indiv_P2"].values()))
    results["total_num_blobs"] = (
        results["num_indiv_gt_blobs"] + results["num_crossing_gt_blobs"]
    )
    results["total_indiv_blobs"] = np.sum(list(results["num_blobs"].values()))
    results["total_num_errors"] = np.sum(list(results["errors_blobs"].values()))
    results["total_assigned_blobs"] = np.sum(list(results["num_id_blobs"].values()))
    results["total_errors_assigned_blobs"] = np.sum(
        list(results["errors_id_blobs"].values())
    )
    results["total_id_blobs_accum"] = np.sum(
        list(results["num_id_blobs_accum"].values())
    )
    results["total_errors_accum"] = np.sum(
        list(results["errors_id_blobs_accum"].values())
    )
    results["total_id_blobs_after_accum"] = np.sum(
        list(results["num_blobs_after_accum"].values())
    )
    results["total_errors_after_accum"] = np.sum(
        list(results["errors_blobs_after_accum"].values())
    )


def compare_tracking_with_ground_truth(
    num_animals, gt_blobs_in_video, blobs_in_video, ids_perm_dict
):
    """
    This function only considers individual blobs

    :param num_animals:
    :param gt_blobs_in_video:
    :param blobs_in_video:
    :param ids_perm_dict:
    :return:
    """
    # create dictionary to store counters
    results = init_results_dict(num_animals)
    both_blobs_in_video = zip(gt_blobs_in_video, blobs_in_video)
    for gt_blobs_in_frame, blobs_in_frame in both_blobs_in_video:
        results = compare_frame(
            results, blobs_in_frame, gt_blobs_in_frame, ids_perm_dict
        )
    aggregate_counters(results)
    return results


def check_gt_video_consistency(video, gt_video):
    """Checks that the `video` and `gt_video` are consistent: they have
    the same number of frames, the same number animals.

    :param video: <Video object> of the tracked video
    :param gt_video: <Video object> of the tracking session from which the
    ground_truth file was computed
    """
    if video.number_of_frames != gt_video.number_of_frames:
        raise ValueError(
            "The number of frames in the video and in the ground truth video "
            "are different. The ground truth file cannot be reused"
        )

    if video.number_of_animals != gt_video.number_of_animals:
        raise ValueError(
            "The number of animals in the video and in the ground truth video "
            "are different. The ground truth file cannot be reused"
        )

    video_frame_dims = (video.original_height, video.original_width)
    gt_video_frame_dims = (gt_video.original_height, gt_video.original_width)
    if video_frame_dims != gt_video_frame_dims:
        raise ValueError(
            "The video frame dimensions of the ground truth video"
            "and the new video are different"
        )


def get_ids_perm_dict(gt_blobs_in_frame, blobs_in_frame):
    ids_perm_dict = {}
    for blob in blobs_in_frame:
        corresponding_blobs = get_corresponding_gt_blob(blob, gt_blobs_in_frame)
        if len(corresponding_blobs) == 1:
            ids_perm_dict[corresponding_blobs[0].gt_identity] = (
                blob.assigned_identities[0]
            )
        else:
            break
    return ids_perm_dict


def get_permutation_of_identities(
    video, fff_global_fragment, gt_blobs_in_video, blobs_in_video
):
    """Returns a dictionary with the permutation of identities to be
    considered when comparing identities of the ground truth data with the
    new trajectories

    :param video: <Video object> of the new video
    :param fff_global_fragment: int indicating the first frame
    of the core of the first global fragment of the new video
    :param gt_blobs_in_video: list of lists of <GroundTruthBlob
    object> with the ground truth blobs of each frame
    :param blobs_in_video: list of lists of <Blob object> with the blobs
    of each frame of the new video
    :return: dict, key is identity in gt data, value is identity in new video
    """
    ids_perm_dict = {}
    permutation_found = False
    while not permutation_found:
        gt_blobs_in_frame = gt_blobs_in_video[fff_global_fragment]
        blobs_in_frame = blobs_in_video[fff_global_fragment]

        if len(gt_blobs_in_frame) == video.number_of_animals:
            ids_perm_dict = get_ids_perm_dict(gt_blobs_in_frame, blobs_in_frame)
            if len(ids_perm_dict) == video.number_of_animals:
                permutation_found = True
            else:
                fff_global_fragment += 1
        else:
            fff_global_fragment += 1

        if fff_global_fragment > video.number_of_frames:
            raise Exception("No identities permutation found")
    logging.info(f"The identities permutation is {ids_perm_dict}")
    assert len(ids_perm_dict) == video.number_of_animals
    return ids_perm_dict


def compute_performance(results, number_of_animals):
    accuracies = {
        "percentage_of_unoccluded_images": (
            results["num_indiv_gt_blobs"] / results["total_num_blobs"]
        ),
        "mean_individual_P2_in_validated_part": (
            results["total_sum_P2"] / results["total_indiv_blobs"]
        ),
    }

    error_rate = results["total_num_errors"] / results["total_indiv_blobs"]
    accuracies["accuracy"] = 1.0 - error_rate

    error_rate = (
        results["total_errors_assigned_blobs"] / results["total_assigned_blobs"]
    )
    accuracies["accuracy_assigned"] = 1.0 - error_rate

    error_rate = results["total_errors_accum"] / results["total_id_blobs_accum"]
    accuracies["accuracy_in_accumulation"] = 1.0 - error_rate

    if results["total_id_blobs_after_accum"] != 0:
        error_rate = (
            results["total_errors_after_accum"] / results["total_id_blobs_after_accum"]
        )
        accuracies["accuracy_after_accumulation"] = 1.0 - error_rate
    else:
        accuracies["accuracy_after_accumulation"] = None

    if results["num_crossing_gt_blobs"] != 0:
        correct = results["crossing_detector_tn"] + results["crossing_detector_tp"]
        positive = results["crossing_detector_tp"] + results["crossing_detector_fp"]
        negative = results["crossing_detector_tn"] + results["crossing_detector_fn"]
        total = positive + negative
        accuracies["crossing_detector_accuracy"] = correct / total
        accuracies["crossing_detector_precision"] = (
            results["crossing_detector_tp"] / positive
        )
        accuracies["crossing_detector_recall"] = results["crossing_detector_tp"] / (
            results["crossing_detector_tp"] + results["crossing_detector_fn"]
        )
    else:
        accuracies["crossing_detector_accuracy"] = None

    accuracies["individual_P2_in_validated_part"] = {}
    accuracies["individual_accuracy"] = {}
    accuracies["individual_accuracy_assigned"] = {}
    accuracies["individual_accuracy_in_accumulation"] = {}
    accuracies["individual_accuracy_after_accumulation"] = {}
    for i in range(1, number_of_animals + 1):
        accuracies["individual_P2_in_validated_part"][i] = (
            results["sum_indiv_P2"][i] / results["num_blobs"][i]
        )
        if results["num_blobs"] != 0:
            error_rate = results["errors_blobs"][i] / results["num_blobs"][i]
            accuracies["individual_accuracy"][i] = 1 - error_rate
        else:
            accuracies["individual_accuracy"][i] = None

        if results["num_id_blobs"] != 0:
            error_rate = results["errors_id_blobs"][i] / results["num_id_blobs"][i]
            accuracies["individual_accuracy_assigned"][i] = 1 - error_rate
        else:
            accuracies["individual_accuracy_assigned"][i] = None

        if results["num_id_blobs_accum"][i] != 0:
            error_rate = (
                results["errors_id_blobs_accum"][i] / results["num_id_blobs_accum"][i]
            )
            accuracies["individual_accuracy_in_accumulation"][i] = 1 - error_rate
        else:
            accuracies["individual_accuracy_in_accumulation"][i] = None

        if results["num_blobs_after_accum"][i] != 0:
            error_rate = (
                results["errors_blobs_after_accum"][i]
                / results["num_blobs_after_accum"][i]
            )
            accuracies["individual_accuracy_after_accumulation"][i] = 1 - error_rate
        else:
            accuracies["individual_accuracy_after_accumulation"][i] = None

    logging.info("accuracies %s" % str(accuracies))
    return accuracies


def get_accuracy_wrt_groundtruth(
    video,
    gt_blobs_in_video,
    blobs_in_video=None,
    identities_dictionary_permutation=None,
):
    number_of_animals = video.number_of_animals
    if blobs_in_video is None:
        blobs_in_video = gt_blobs_in_video

    results = compare_tracking_with_ground_truth(
        number_of_animals,
        gt_blobs_in_video,
        blobs_in_video,
        identities_dictionary_permutation,
    )
    pprint(results)

    if len(results["frames_w_0_id_in_gt"]) == 0:
        accuracies = compute_performance(results, number_of_animals)
        return accuracies, results

    logging.info(
        "there are fish with 0 identity in frame %s",
        str(results["frames_w_0_id_in_gt"]),
    )
    return None, results


def reduce_pixels(
    blob, original_width, original_height, width, height, resolution_reduction
):
    x, y = np.unravel_index(blob.pixels, (original_height, original_width))
    x = np.round(x * resolution_reduction).astype(int)
    y = np.round(y * resolution_reduction).astype(int)
    return np.ravel_multi_index([x, y], (height, width))


def reduce_resolution_gt_blobs(video, gt_blobs_in_video):
    for gt_blobs_in_frame in gt_blobs_in_video:
        for gt_blob in gt_blobs_in_frame:
            gt_blob.pixels = reduce_pixels(
                gt_blob,
                video.original_width,
                video.original_height,
                video.width,
                video.height,
                video.resolution_reduction,
            )


def compute_and_save_session_accuracy_wrt_groundtruth(video: Video, gt_type=None):
    if gt_type == "normal":
        list_of_blobs_path = video.blobs_path
        performance_func = get_accuracy_wrt_groundtruth
    elif gt_type == "no_gaps":
        raise Exception(f"No performance_func to compute for {gt_type}")
        # TODO: fixh get_accuracy_wrt_groundtruth_no_gaps
        # list_of_blobs_path = video.blobs_no_gaps_path
        # gt_path = os.path.join(
        #     video.video_folder, "_groundtruth_with_crossing_identified.npy"
        # )
        # performance_func = get_accuracy_wrt_groundtruth_no_gaps

    else:
        raise ValueError(f"Not valid gt_type {gt_type}")

    logging.info("loading list_of_blobs")
    list_of_blobs = ListOfBlobs.load(list_of_blobs_path)

    logging.info("loading ground truth")
    ground_truth = np.load(
        video.ground_truth_path, allow_pickle=True, encoding="latin1"
    ).item()

    check_gt_video_consistency(video, ground_truth.video)

    if video.resolution_reduction != 1:
        reduce_resolution_gt_blobs(video, ground_truth.blobs_in_video)

    accumulation_number = int(video.accumulation_folder[-1])
    identities_dictionary_permutation = get_permutation_of_identities(
        video,
        video.first_frame_first_global_fragment[accumulation_number],
        ground_truth.blobs_in_video,
        list_of_blobs.blobs_in_video,
    )

    # Select the frame for which we checked the identities and computed the
    # ground_truth file
    gt_blobs_in_video = ground_truth.blobs_in_video[
        ground_truth.start : ground_truth.end
    ]

    blobs_in_video = list_of_blobs.blobs_in_video[ground_truth.start : ground_truth.end]

    logging.info("computing performance")
    accuracies, results = performance_func(
        video, gt_blobs_in_video, blobs_in_video, identities_dictionary_permutation
    )

    if accuracies is not None:
        save_accuracies_in_video(
            video, accuracies, results, (ground_truth.start, ground_truth.end), gt_type
        )

    return video, ground_truth


def save_accuracies_in_video(video, accuracies, results, gt_start_end, gt_type):
    logging.info("saving accuracies in video")
    video.gt_start_end = gt_start_end
    if gt_type == "normal":
        video.gt_accuracy = accuracies
        video.gt_results = results
    elif gt_type == "no_gaps":
        video.gt_accuracy_no_gaps = accuracies
        video.gt_results_no_gaps = results
    video.save()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-gt",
        "--gt_type",
        type=str,
        help="type of ground_truth to compute ('no_gaps' or 'normal')",
    )
    parser.add_argument(
        "-sf", "--session_folder", type=str, help="path to the session folder"
    )
    args = parser.parse_args()

    gt_type = args.gt_type
    session_folder = args.session_folder
    video_object_path = os.path.join(session_folder, "video_object.npy")
    logging.info("loading video object")
    video = np.load(video_object_path, allow_pickle=True, encoding="latin1").item(0)
    video.update_paths(video_object_path)
    ground_truth = np.load(
        video.ground_truth_path, allow_pickle=True, encoding="latin1"
    ).item()
    video, ground_truth = compute_and_save_session_accuracy_wrt_groundtruth(
        video, gt_type
    )
    print(video.gt_accuracy)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import json
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from math import sqrt
from pathlib import Path
from shutil import rmtree
from typing import Iterable, Optional, TypeVar

import cv2
import h5py
import numpy as np
from rich.progress import BarColumn, Progress, TaskProgressColumn, TimeRemainingColumn

from .init_logger import CustomError

InputType = TypeVar("InputType")


def track(
    sequence: Iterable[InputType],  # TODO also Sequence?
    desc: str = "Working...",
    total: Optional[float] = None,
) -> Iterable[InputType]:
    """A custom interpretation of rich.progress.track"""

    progress = Progress(
        " " * 18 + desc,
        BarColumn(bar_width=None),
        TaskProgressColumn(show_speed=True),
        TimeRemainingColumn(elapsed_when_finished=True),
        transient=True,
    )

    with progress:
        yield from progress.track(sequence, total, description=desc)

    task = progress.tasks[0]

    logging.info(
        "[green]%s[/] (%s iterations). It took %s",
        desc,
        int(task.total) if task.total is not None else "unknown",
        "--:--" if task.elapsed is None else timedelta(seconds=int(task.elapsed)),
        stacklevel=3,
        extra={"markup": True},
    )


def delete_attributes_from_object(object_to_modify, list_of_attributes):
    for attribute in list_of_attributes:
        if hasattr(object_to_modify, attribute):
            delattr(object_to_modify, attribute)


def create_dir(path: Path, remove_existing=False):
    if path.is_dir():
        if remove_existing:
            rmtree(path)
            path.mkdir()
            logging.info(f"Directory {path} has been cleaned", stacklevel=3)
        else:
            logging.info(f"Directory {path} already exists", stacklevel=3)
    else:
        if not path.parent.is_dir():
            path.parent.mkdir()
        path.mkdir()
        logging.info(f"Directory {path} has been created", stacklevel=3)


def remove_dir(path: Path):
    if path.is_dir():
        rmtree(path, ignore_errors=True)
        logging.info(f"Directory {path} has been removed", stacklevel=3)
    else:
        logging.info(f"Directory {path} not found, can't remove", stacklevel=3)


def remove_file(path: Path):
    if path.is_file():
        path.unlink()
        logging.info(f"File {path} has been removed", stacklevel=3)


def assert_all_files_exist(paths: list[Path]):
    """Returns FileNotFoundError if any of the paths is not an existing file"""
    for path in paths:
        if not path.is_file():
            raise FileNotFoundError(f"File {path} not found")


def get_vertices_from_label(label: str, close=False):
    """Transforms a string representation of a polygon from the
    ROI widget (idtrackerai_app) into a vertices np.array"""
    try:
        data = json.loads(label[10:].replace("'", '"'))
    except ValueError:
        raise CustomError(f'Not recognized ROI representation: "{label}"')

    if label[2:9] == "Polygon":
        vertices = np.asarray(data)
    elif label[2:9] == "Ellipse":
        vertices = cv2.ellipse2Poly(
            data["center"], data["axes"], data["angle"], 0, 360, 2
        )
    else:
        raise TypeError(label)

    if close:
        return np.vstack([vertices, vertices[0]])
    return vertices


def build_ROI_mask_from_list(
    list_of_ROIs: None | list[str] | str,
    resolution_reduction: float,
    width: int,
    height: int,
) -> np.ndarray | None:
    """Transforms a list of polygons (as type str) from
    ROI widget (idtrackerai_app) into a boolean np.array mask"""

    if list_of_ROIs is None:
        return None
    ROI_mask = np.zeros(
        (
            int(height * resolution_reduction + 0.5),
            int(width * resolution_reduction + 0.5),
        ),
        np.uint8,
    )

    if isinstance(list_of_ROIs, str):
        list_of_ROIs = [list_of_ROIs]

    for line in list_of_ROIs:
        vertices = (get_vertices_from_label(line) * resolution_reduction + 0.5).astype(
            np.int32
        )
        if line[0] == "+":
            cv2.fillPoly(ROI_mask, (vertices,), color=255)
        elif line[0] == "-":
            cv2.fillPoly(ROI_mask, (vertices,), color=0)
        else:
            raise TypeError
    return ROI_mask


@dataclass(slots=True)
class Episode:
    index: int
    local_start: int
    local_end: int
    video_path_index: int
    global_start: int
    global_end: int


class Timer:
    """Simple class for measuring execution time during the whole process"""

    start_time: datetime | None = None
    finish_time: datetime | None = None

    def __init__(self, name: str = ""):
        self.name = name

    def reset(self):
        self.start_time = None
        self.finish_time = None

    @property
    def interval(self):
        if self.finish_time is None or self.start_time is None:
            return None
        return self.finish_time - self.start_time

    @property
    def started(self):
        return self.start_time is not None

    @property
    def finished(self):
        return self.interval is not None

    def start(self):
        logging.info(
            "[blue bold]START %s", self.name, extra={"markup": True}, stacklevel=3
        )
        self.start_time = datetime.now()

    def finish(self, raise_if_not_started=True):
        if not self.started and raise_if_not_started:
            raise RuntimeError("Timer finish method called before start method")

        self.finish_time = datetime.now()

        logging.info(
            f"[blue bold]FINISH {self.name}, it took {self}",
            extra={"markup": True},
            stacklevel=3,
        )

    def __str__(self) -> str:
        return str(self.interval or "Not finished").split(".")[0]

    @classmethod
    def from_dict(cls, d: dict):
        obj = cls.__new__(cls)
        obj.name = d["name"]

        if "interval" in d:  # v5.1.0 compatibility
            if d["start_time"] > 0:
                obj.start_time = datetime.fromtimestamp(d["start_time"])

            if d["interval"] > 0:
                obj.finish_time = datetime.fromtimestamp(
                    d["start_time"] + d["interval"]
                )

        else:
            if "start_time" in d:
                obj.start_time = datetime.fromisoformat(d["start_time"])
            if "finish_time" in d:
                obj.finish_time = datetime.fromisoformat(d["finish_time"])

        return obj


def assert_knowledge_transfer_is_possible(
    knowledge_transfer_folder: Path | None, n_animals: int
) -> list[int]:
    if knowledge_transfer_folder is None:
        raise CustomError(
            "To perform knowledge/identity transfer you "
            "need to provide a path for the variable "
            "'KNOWLEDGE_TRANSFER_FOLDER'"
        )

    model_params_path = knowledge_transfer_folder / "model_params.json"
    if model_params_path.is_file():
        model_params_dict = json.load(model_params_path.open())
        n_classes, image_size = extract_parameters_from_model_json(model_params_dict)

    elif model_params_path.with_suffix(".npy").is_file():
        model_params_dict = np.load(
            model_params_path.with_suffix(".npy"), allow_pickle=True
        ).item()  # loading from v4
        n_classes, image_size = extract_parameters_from_model_json(model_params_dict)

    else:
        logging.warning('"%s" file not found', model_params_path)
        n_classes, image_size = extract_parameters_from_model_state_dict(
            knowledge_transfer_folder
        )

    if n_animals != n_classes:
        raise CustomError(
            "Tracking with knowledge/identity transfer is not possible. "
            "The number of animals in the video needs to be the same as "
            "the number of animals in the transferred network."
        )

    logging.info(
        "Tracking with knowledge transfer. "
        "The identification image size will be matched "
        "to the image_size of the transferred network: %s",
        image_size,
    )
    return image_size


def extract_parameters_from_model_json(model_parameters: dict):
    image_size = model_parameters["image_size"]
    n_classes = (
        model_parameters["n_classes"]
        if "n_classes" in model_parameters  # 5.1.6 compatibility
        else model_parameters["number_of_classes"]
    )
    return n_classes, image_size


def extract_parameters_from_model_state_dict(knowledge_transfer_folder: Path):
    logging.info("Extracting model parameters from state dictionary")
    # this import is here (not at the top of the file) to avoid its loading process
    # when loading GUIs without identity_transfer (almost always)
    import torch

    model_dict_path = knowledge_transfer_folder / "identification_network.model.pth"
    model_state_dict: dict[str, torch.Tensor] = torch.load(model_dict_path)
    if "fc2.weight" in model_state_dict:
        layer_in_dimension = model_state_dict["fc1.weight"].size(1)
        n_classes = len(model_state_dict["fc2.weight"])
    else:
        layer_in_dimension = model_state_dict["layers.9.weight"].size(1)
        n_classes = len(model_state_dict["layers.11.weight"])
    image_size = int(4 * sqrt(layer_in_dimension / 100)) + 2
    return n_classes, [image_size, image_size, 1]


def pprint_dict(d: dict, name: str = "") -> str:
    text = f"[bold blue]{name}[/]:" if name else ""

    pad = min(max(map(len, d.keys())), 25)

    for key, value in d.items():
        if isinstance(value, tuple):
            value = list(value)
        if isinstance(value, list) and value and isinstance(value[0], Path):
            value = list(map(str, value))
        if isinstance(value, Path):
            value = str(value)
        if len(repr(value)) < 50 or not isinstance(value, list):
            text += f"\n[bold]{key:>{pad}}[/] = {repr(value)}"
        else:
            s = f"[{repr(value[0])}"
            for item in value[1:]:
                s += f",\n{' '*pad}    {repr(item)}"
            s += "]"
            text += f"\n[bold]{key:>{pad}}[/] = {s}"
    return text


def load_id_images(
    id_images_file_paths: list[Path], images_indices: Iterable[tuple[int, int]]
) -> np.ndarray:
    """Loads the identification images from disk.

    Parameters
    ----------
    id_images_file_paths : list
        List of strings with the paths to the files where the images are
        stored.
    images_indices : list
        List of tuples (image_index, episode) that indicate each of the images
        to be loaded

    Returns
    -------
    Numpy array
        Numpy array of shape [number of images, width, height]
    """
    if isinstance(images_indices, zip):
        images_indices = list(images_indices)

    img_indices, episodes = np.asarray(images_indices).T

    # Create entire output array
    with h5py.File(id_images_file_paths[0], "r") as file:
        test_dataset = file["id_images"]
        images = np.empty(
            (len(images_indices), *test_dataset.shape[1:]), test_dataset.dtype  # type: ignore
        )

    for episode in track(set(episodes), "Loading identification images from disk"):
        where = episodes == episode
        with h5py.File(id_images_file_paths[episode], "r") as file:
            images[where] = file["id_images"][:][img_indices[where]]  # type: ignore

    return images


def json_default(obj):
    """Encodes non JSON serializable object as dicts"""
    if isinstance(obj, Path):
        return {"py/object": "Path", "path": str(obj)}

    if isinstance(obj, (Timer, Episode)):
        return {"py/object": obj.__class__.__name__} | obj.__dict__

    if isinstance(obj, np.integer):
        return int(obj)

    if isinstance(obj, np.floating):
        return float(obj)

    if isinstance(obj, np.ndarray):
        return {"py/object": "np.ndarray", "values": obj.tolist()}

    if isinstance(obj, set):
        return {"py/object": "set", "values": list(obj)}

    if isinstance(obj, datetime):
        return obj.isoformat()

    raise ValueError(f"Could not JSON serialize {obj} of type {type(obj)}")


def json_object_hook(d: dict):
    """Decodes dicts from `json_default`"""
    if "py/object" in d:
        cls = d.pop("py/object")
        if cls == "Path":
            return Path(d["path"])
        if cls == "Episode":
            return Episode(**d)
        if cls == "Timer":
            return Timer.from_dict(d)
        if cls == "np.ndarray":
            return np.asarray(d["values"])
        if cls == "set":
            return set(d["values"])
        raise ValueError(f"Could not read {d}")
    return d


def resolve_path(path: Path | str) -> Path:
    return Path(path).expanduser().resolve()


def clean_attrs(obj: object):
    """Removes instances attributes if they are redundant
    with the class attributes"""
    class_attr = obj.__class__.__dict__

    attributes_to_remove: list[str] = [
        attr
        for attr, value in obj.__dict__.items()
        if attr in class_attr
        and isinstance(class_attr[attr], type(value))
        and class_attr[attr] == value
    ]

    for attr in attributes_to_remove:
        delattr(obj, attr)



import logging
from pathlib import Path

import numpy as np

from idtrackerai import Video
from idtrackerai.utils import wrap_entrypoint

from .general_video import generate_trajectories_video
from .individual_videos import generate_individual_video


@wrap_entrypoint
def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "session_path",
        type=Path,
        help="Path to the video session created during the tracking session",
        metavar="",
    )

    parser.add_argument(
        "--individual",
        action="store_true",
        help="Generate individual video. Default is a general video",
    )
    parser.add_argument(
        "--gray", action="store_true", help="Draw the original video in grayscale"
    )

    parser.add_argument(
        "--t",
        type=Path,
        help=(
            "Path to the trajectory file, default is "
            "session_dir/trajectories/without_gaps.npy"
        ),
        metavar="",
    )
    parser.add_argument(
        "--tl",
        type=int,
        default=20,
        help=(
            "Trail length, number of points used to draw the individual trajectories"
            " traces"
        ),
        metavar="",
    )
    parser.add_argument(
        "--s", type=int, default=0, help="Frame where to start the video", metavar=""
    )
    parser.add_argument(
        "--e", type=int, help="Frame where to end the video", metavar=""
    )
    args = parser.parse_args()

    video = Video.load(args.session_path)

    if args.t is None:
        possible_files = (
            "validated.npy",
            "without_gaps.npy",
            "with_gaps.npy",
            "trajectories_validated.npy",
            "trajectories_wo_gaps.npy",
            "trajectories.npy",
            "trajectories_wo_identification.npy",
        )
        for file in possible_files:
            path = video.trajectories_folder / file
            if path.is_file():
                logging.info("Loading trajectories from %s", path)
                trajectories = np.load(path, allow_pickle=True).item()["trajectories"]
                break
        else:
            raise FileNotFoundError(
                f"Could not find the trajectory file in {video.trajectories_folder}"
            )
    else:
        logging.info("Loading trajectories from %s", args.t)
        trajectories = np.load(args.t, allow_pickle=True).item()["trajectories"]

    if args.individual:
        generate_individual_video(
            video,
            trajectories,
            draw_in_gray=args.gray,
            starting_frame=args.s,
            ending_frame=args.e,
        )
    else:
        generate_trajectories_video(
            video,
            trajectories,
            draw_in_gray=args.gray,
            centroid_trace_length=args.tl,
            starting_frame=args.s,
            ending_frame=args.e,
        )


if __name__ == "__main__":
    main()



from re import compile

from qtpy.QtCore import Qt, Signal  # type: ignore
from qtpy.QtGui import QColor, QColorConstants
from qtpy.QtWidgets import QInputDialog, QToolButton, QVBoxLayout, QWidget

from idtrackerai_GUI_tools import CanvasMouseEvent, CanvasPainter, CustomList


def has_invalid_chars(string):
    if len(string) == 0:
        return True
    regex = compile(r"[@!$%^&*?/\~:|]")
    return regex.search(string) is not None


QColors = [
    QColor(0x9467BD),
    QColor(0x2CA02C),
    QColor(0xBCBD22),
    QColor(0xFF7F0E),
    QColor(0x8C564B),
    QColor(0xE377C2),
    QColor(0x7F7F7F),
    QColor(0x17BECF),
]
n_colors = len(QColors)


class SetupPoints(QWidget):
    needToDraw = Signal()

    def __init__(self):
        super().__init__()

        self.add = QToolButton()
        self.add.setText("Add")
        self.add.setCheckable(True)

        self.list = CustomList(max_n_row=0)

        layout = QVBoxLayout()
        layout.setSpacing(2)
        self.setLayout(layout)
        layout.addWidget(self.add)
        layout.addWidget(self.list)

        self.add.toggled.connect(self.add_clicked)
        self.setup_points_dict: dict[str, tuple[QColor, list[tuple[int, int]]]] = {}
        self.list.ListChanged.connect(self.needToDraw.emit)
        self.list.removedItem.connect(self.remove_item)
        self.color_count = -1
        self.setup_name = None

    def click_event(self, event: CanvasMouseEvent):
        if self.isVisible() and self.isEnabled() and self.setup_name is not None:
            points = self.setup_points_dict[self.setup_name][1]
            if event.button == Qt.MouseButton.LeftButton:
                # Add clicked point
                points.append(event.int_xy_data)
            elif event.button == Qt.MouseButton.RightButton:
                # Remove nearest point
                if not points:
                    return
                distances = map(event.distance_to, points)
                index, dist = min(enumerate(distances), key=lambda x: x[1])
                if dist < 20 * event.zoom:  # 20 px threshold
                    points.pop(index)
            self.needToDraw.emit()

    def add_clicked(self, checked):
        if checked:
            existing_names = [
                self.list.item(i).data(Qt.ItemDataRole.UserRole).split(":")[0]
                for i in range(self.list.count())
            ]
            invalid = True
            dialog_text = "Enter setup points name:"
            while invalid:
                name, ok = QInputDialog.getText(self.add, "idtracker.ai", dialog_text)
                if not ok:
                    self.add.setChecked(False)
                    return
                name = name.strip()
                invalid = has_invalid_chars(name)
                if invalid:
                    dialog_text = "Invalid characters encountered"
                if name in existing_names:
                    invalid = True
                    dialog_text = "Enter a non existing name"
            self.color_count = (
                0 if self.color_count == n_colors - 1 else self.color_count + 1
            )

            self.setup_name = name
            self.setup_points_dict[name] = (QColors[self.color_count], [])
            self.needToDraw.emit()

        else:
            if self.setup_name is None:
                return
            self.list.add_str(
                self.setup_name
                + ": "
                + ",".join(
                    [
                        f"[{x:d}, {y:d}]"
                        for x, y in self.setup_points_dict[self.setup_name][1]
                    ]
                ),
                color=self.setup_points_dict[self.setup_name][0],
            )
            self.setup_name = None

    def remove_item(self, data: str):
        self.setup_points_dict.pop(data.split(":")[0])

    def load_points(self, values: dict[str, list[tuple[int, int]]]):
        self.list.clear()
        self.setup_points_dict.clear()
        if not isinstance(values, dict):
            return  # Loading from v4

        for name, points in values.items():
            self.color_count = (
                0 if self.color_count == n_colors - 1 else self.color_count + 1
            )
            self.setup_points_dict[name] = (QColors[self.color_count], points)
            self.list.add_str(
                name
                + ": "
                + ",".join([f"[{int(x):d}, {int(y):d}]" for x, y in points]),
                color=QColors[self.color_count],
            )

    def get_points(self) -> dict[str, list[tuple[int, int]]]:
        return {key: value[1] for key, value in self.setup_points_dict.items()}

    def paint_on_canvas(self, painter: CanvasPainter):
        painter.setPenColor(QColorConstants.Black)
        for color, points in self.setup_points_dict.values():
            painter.setBrush(color)
            for point in points:
                painter.drawBigPoint(*point)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import json
import logging
from argparse import ArgumentParser
from pathlib import Path

import numpy as np

from idtrackerai.utils import create_dir, resolve_path, wrap_entrypoint


def save_array_to_csv(path: Path, array: np.ndarray, key: str, fps=float | None):
    array = array.squeeze()
    if key == "id_probabilities":
        fmt = "%.3e"
    elif key == "trajectories":
        fmt = "%.3f"
    else:
        fmt = "%.3f"

    if array.ndim == 3:
        array_header = ",".join(
            coord + str(i) for i in range(1, array.shape[1] + 1) for coord in ("x", "y")
        )
        array = array.reshape((-1, array.shape[1] * array.shape[2]))
    elif array.ndim == 2:
        array_header = ",".join(f"{key}{i}" for i in range(1, array.shape[1] + 1))
    else:
        raise ValueError(array.shape)

    fmt = [fmt] * array.shape[1]

    if fps is not None:  # add time column
        array_header = "seconds," + array_header
        fmt = ["%.3f"] + fmt
        time = np.arange(array.shape[0], dtype=float) / fps
        array = np.column_stack((time, array))

    np.savetxt(path, array, delimiter=",", header=array_header, fmt=fmt, comments="")


def convert_trajectories_file_to_csv_and_json(
    npy_path: Path, add_time_column: bool = False, raise_errors=False
):
    logging.info(f"Converting {npy_path} to .csv and .json")
    output_dir = npy_path.parent / (npy_path.stem + "_csv")
    create_dir(output_dir, remove_existing=True)
    try:
        trajectories_dict: dict = np.load(npy_path, allow_pickle=True).item()
        attributes_dict = {}
        for key, value in trajectories_dict.items():
            if key in ("trajectories", "id_probabilities"):
                save_array_to_csv(
                    output_dir / (key + ".csv"),
                    value,
                    key=key,
                    fps=(
                        trajectories_dict.get("frames_per_second", 1)
                        if add_time_column
                        else None
                    ),
                )
            elif key == "areas":
                np.savetxt(
                    output_dir / (key + ".csv"),
                    np.asarray((value["mean"], value["median"], value["std"])).T,
                    delimiter=",",
                    header="mean, median, standard_deviation",
                    fmt="%.1f",
                    comments="",
                )
            else:
                attributes_dict[key] = value

        json.dump(attributes_dict, (output_dir / "attributes.json").open("w"), indent=4)
    except Exception as e:
        if raise_errors:
            raise e
        logging.error(e)


@wrap_entrypoint
def main():
    parser = ArgumentParser()

    parser.add_argument(
        "paths",
        help=(
            "Paths to convert trajectories to CSV and JSON. Can be session folders (to"
            " convert all .npy files inside trajectory subfolder), arbitrary folder (to"
            " convert all .npy files in it) and specific .npy files."
        ),
        type=Path,
        nargs="+",
    )
    parser.add_argument(
        "--add_time",
        help="Add a time column (in seconds) to csv trajectory files.",
        action="store_true",
    )

    args = parser.parse_args()

    for path in args.paths:
        path = resolve_path(path)
        if not path.exists():
            logging.warning('Path "%s" not found', path)
            continue
        files_found = False
        if path.is_file() and path.suffix == ".npy":
            convert_trajectories_file_to_csv_and_json(
                path, args.add_time, raise_errors=True
            )
            files_found = True

        if path.name.startswith("session_"):
            path /= "trajectories"

        for file in path.glob("*.npy"):
            convert_trajectories_file_to_csv_and_json(
                file, args.add_time, raise_errors=True
            )
            files_found = True

        if not files_found:
            logging.warning('No trajectory files found in "%s"', path)


if __name__ == "__main__":
    main()



import logging
from dataclasses import dataclass
from math import sqrt

from qtpy.QtCore import QPoint, QPointF, Qt, Signal  # type: ignore
from qtpy.QtGui import (
    QColor,
    QColorConstants,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QPolygon,
    QWheelEvent,
)
from qtpy.QtWidgets import QWidget


@dataclass(slots=True)
class CanvasMouseEvent:
    button: Qt.MouseButton
    """Clicked button"""
    zoom: float
    """Current zoom of the canvas when event was created"""
    xy_data: tuple[float, float]
    """Position of the click in content data units"""

    def distance_to(self, point: tuple[float, float]):
        return sqrt(self.sq_distance_to(point))

    def sq_distance_to(self, point: tuple[float, float]):
        return (point[0] - self.xy_data[0]) ** 2 + (point[1] - self.xy_data[1]) ** 2

    @property
    def int_xy_data(self):
        return int(self.xy_data[0] + 0.5), int(self.xy_data[1] + 0.5)


class CanvasPainter(QPainter):
    def __init__(self, parent, zoom: float):
        self.applied_zoom = zoom
        super().__init__(parent)

    def drawPolygonFromVertices(self, vertices, scale: float):
        poly = QPolygon()
        poly.setPoints(
            *[int(coord * scale + 0.5) for point in vertices for coord in point]
        )
        super().drawPolygon(poly)

    def setPenColor(self, color: QColor | int | Qt.GlobalColor):
        super().setPen(color)
        pen = self.pen()
        pen.setWidthF(1.3 * self.applied_zoom)
        super().setPen(pen)

    def drawBigPoint(self, x: float, y: float, size=7):
        radi = (size * self.applied_zoom) / 2
        super().drawEllipse(QPointF(x, y), radi, radi)


class Canvas(QWidget):
    # TODO check better implementations with
    # QGraphicsItem, QGraphicsScene, QtQuick, Canvas

    click_event = Signal(CanvasMouseEvent)
    double_click_event = Signal(CanvasMouseEvent)
    painting_time = Signal(CanvasPainter)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.zoom = 3.0
        self.centerX: float = 0.0
        self.centerY: float = 0.0
        self.has_moved: bool = False
        self.mouse_pressed: bool = False
        self.click_origin: tuple = (0, 0)
        self.real_w_zoom: float
        self.real_h_zoom: float
        self.real_x0: int
        self.real_y0: int
        self.setCursor(Qt.CursorShape.PointingHandCursor)

    def paintEvent(self, event: QPaintEvent):
        painter = CanvasPainter(self, self.zoom)
        try:
            painter_rect = event.rect()
            painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
            painter.fillRect(painter_rect, QColorConstants.Black)
            axis_w = painter_rect.width() * self.zoom
            axis_h = painter_rect.height() * self.zoom

            # save inaccuracies in rounding to use in self.to_physical_units
            self.real_w_zoom = int(axis_w) / painter_rect.width()
            self.real_h_zoom = int(axis_h) / painter_rect.height()
            self.real_x0 = int(self.centerX - axis_w / 2)
            self.real_y0 = int(self.centerY - axis_h / 2)

            painter.setWindow(self.real_x0, self.real_y0, int(axis_w), int(axis_h))

            font = self.font()
            font.setPointSizeF(font.pointSizeF() * 1.3 * self.zoom)
            painter.setFont(font)

            pen = painter.pen()
            pen.setWidthF(1.8 * self.zoom)
            painter.setPen(pen)

            self.painting_time.emit(painter)
        except Exception as e:
            logging.error(e)

    def to_physical_units(self, point: QPoint | QPointF):
        return (
            self.real_x0 + self.real_w_zoom * point.x(),
            self.real_y0 + self.real_h_zoom * point.y(),
        )

    def wheelEvent(self, event: QWheelEvent):
        step = event.angleDelta().y() / 1200
        if step > 0 and self.zoom < 0.1:
            return
        if step < 0 and self.zoom > 100:
            return
        xdata, ydata = self.to_physical_units(event.position())
        self.centerX += (xdata - self.centerX) * step
        self.centerY += (ydata - self.centerY) * step
        self.zoom *= 1 - step
        self.update()

    def mousePressEvent(self, event: QMouseEvent):
        self.has_moved = False
        self.mouse_pressed = True
        self.click_origin = (event.pos().x(), event.pos().y())

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        self.double_click_event.emit(
            CanvasMouseEvent(
                event.button(), self.zoom, self.to_physical_units(event.pos())
            )
        )

    def mouseReleaseEvent(self, event: QMouseEvent):
        self.mouse_pressed = False
        self.setCursor(Qt.CursorShape.PointingHandCursor)

        if not self.has_moved:
            self.setFocus()
            self.click_event.emit(
                CanvasMouseEvent(
                    event.button(), self.zoom, self.to_physical_units(event.pos())
                )
            )

    def mouseMoveEvent(self, event: QMouseEvent):
        if self.mouse_pressed:
            self.setCursor(Qt.CursorShape.ClosedHandCursor)
            self.has_moved = True

            self.centerX -= self.zoom * (event.pos().x() - self.click_origin[0])
            self.centerY -= self.zoom * (event.pos().y() - self.click_origin[1])
            self.click_origin = (event.pos().x(), event.pos().y())
            self.update()

    def adjust_zoom_to(self, width, height):
        self.centerX = width / 2
        self.centerY = height / 2
        self.zoom = max(width / self.width(), height / self.height())



from .bkg_widget import BkgWidget
from .blob_info_widget import BlobInfoWidget
from .frame_analyzer import FrameAnalyzer
from .intensity_ths import IntensityThresholds
from .open_video_widget import OpenVideoWidget
from .ROI_widget import ROIWidget
from .track_intervals_widget import TrackingIntervalsWidget

__all__ = [
    "BkgWidget",
    "BlobInfoWidget",
    "FrameAnalyzer",
    "IntensityThresholds",
    "OpenVideoWidget",
    "ROIWidget",
    "TrackingIntervalsWidget",
]
